# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

import json
import os

from polyaxon_client import settings
from polyaxon_client.exceptions import PolyaxonClientException
from polyaxon_client.tracking.base import BaseTracker
from polyaxon_client.tracking.in_cluster import ensure_in_custer
from polyaxon_client.tracking.no_op import check_no_op


class Job(BaseTracker):
    @check_no_op
    def __init__(self,
                 project=None,
                 job_id=None,
                 job_type=None,
                 client=None,
                 track_logs=True,
                 track_code=True,
                 track_env=True,
                 outputs_store=None):
        if project is None and settings.IN_CLUSTER and not self.is_notebook_job:
            job_info = self.get_job_info()
            project = job_info['project_name']
            job_id = job_info['job_name'].split('.')[-1]
            job_type = job_info['job_name'].split('.')[-2]

        super(Job, self).__init__(project=project,
                                  client=client,
                                  track_logs=track_logs,
                                  track_code=track_code,
                                  track_env=track_env,
                                  outputs_store=outputs_store)
        self.job_id = job_id
        self.job_type = job_type
        self.job = None
        self.last_status = None

    @check_no_op
    def get_data(self):
        if self.job_type == 'jobs':
            self._data = self.client_backend.get_job(
                username=self.username,
                project_name=self.project_name,
                job_id=self.job_id)
            return
        elif self.job_type == 'builds':
            self._data = self.client_backend.get_build(
                username=self.username,
                project_name=self.project_name,
                job_id=self.job_id)
            return
        raise PolyaxonClientException('Job type {} not supported'.format(self.job_type))

    @property
    def client_backend(self):
        if settings.NO_OP:
            return None

        if self.job_type == 'jobs':
            return self.client.job
        elif self.job_type == 'builds':
            return self.client.build_job
        raise PolyaxonClientException('Job type {} not supported'.format(self.job_type))

    @check_no_op
    def _set_health_url(self):
        health_url = self.client_backend.get_heartbeat_url(
            username=self.username,
            project_name=self.project_name,
            job_id=self.job_id)
        self.client.set_health_check(url=health_url)

    @staticmethod
    def get_job_info():
        """Returns information about the job:
            * project_name
            * job_name
            * project_uuid
            * job_uuid
            * role
            * type
            * app
        """
        if settings.NO_OP:
            return None

        ensure_in_custer()

        info = os.getenv('POLYAXON_JOB_INFO', None)
        try:
            return json.loads(info) if info else None
        except (ValueError, TypeError):
            print('Could get experiment info, '
                  'please make sure this is running inside a polyaxon job.')
            return None

    @check_no_op
    def log_status(self, status, message=None, traceback=None):
        self.client_backend.create_status(username=self.username,
                                          project_name=self.project_name,
                                          job_id=self.job_id,
                                          status=status,
                                          message=message,
                                          traceback=traceback,
                                          background=True)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from polyaxon_client.api.base import BaseApiHandler
from polyaxon_client.exceptions import PolyaxonClientException


class UserApi(BaseApiHandler):
    """
    Api handler to manage users.
    """
    ENDPOINT = "/users"
    ENDPOINT_SUPERUSERS = "/superusers"

    def activate_user(self, username):
        """Activate a user account."""
        request_url = self.build_url(self._get_http_url(), 'activate', username)

        try:
            return self.transport.post(request_url)
        except PolyaxonClientException as e:
            self.transport.handle_exception(e=e, log_message='Error while activating user.')
            return None

    def delete_user(self, username):
        """Delete a user account."""
        request_url = self.build_url(self._get_http_url(), 'delete', username)

        try:
            return self.transport.delete(request_url)
        except PolyaxonClientException as e:
            self.transport.handle_exception(e=e, log_message='Error while deleting user.')
            return None

    def grant_superuser(self, username):
        """Activate a user account."""
        request_url = self.build_url(self._get_http_url(self.ENDPOINT_SUPERUSERS),
                                     'grant',
                                     username)

        try:
            return self.transport.post(request_url)
        except PolyaxonClientException as e:
            self.transport.handle_exception(
                e=e, log_message='Error while grating superuser to user.')
            return None

    def revoke_superuser(self, username):
        """Activate a user account."""
        request_url = self.build_url(self._get_http_url(self.ENDPOINT_SUPERUSERS),
                                     'revoke',
                                     username)

        try:
            return self.transport.post(request_url)
        except PolyaxonClientException as e:
            self.transport.handle_exception(
                e=e, log_message='Error while revoking superuser to user.')
            return None



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

import os

from polyaxon_client.tracking.utils.cmd import run_command


def is_git_initialized(path='.'):
    return bool(run_command(cmd='git rev-parse --is-inside-work-tree',
                            data=None,
                            location=path,
                            chw=True).split('\n')[0])


def get_commit(path='.'):
    return run_command(cmd='git --no-pager log --pretty=oneline -1',
                       data=None,
                       location=path,
                       chw=True).split(' ')[0]


def get_head(path='.'):
    return run_command(cmd='git rev-parse HEAD',
                       data=None,
                       location=path,
                       chw=True).split('\n')[0]


def get_remote(path='.'):
    return run_command(cmd='git config --get remote.origin.url',
                       data=None,
                       location=path,
                       chw=True).split('\n')[0]


def get_repo_name(path='.'):
    repo = run_command(cmd='git rev-parse --show-toplevel',
                       data=None,
                       location=path,
                       chw=True).split('\n')[0]

    return os.path.basename(repo)


def get_branch_name(path='.'):
    return run_command(cmd='git rev-parse --abbrev-ref HEAD',
                       data=None,
                       location=path,
                       chw=True).split('\n')[0]


def is_dirty(path='.'):
    return bool(run_command(cmd='git diff --stat',
                            data=None,
                            location=path,
                            chw=True))


def get_code_reference(path='.'):
    if not is_git_initialized(path):
        return None

    return {
        'commit': get_commit(path),
        'head': get_head(path),
        'branch': get_branch_name(path),
        'git_url': get_remote(path),
        'is_dirty': is_dirty(path),
    }



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from hestia.auth import AuthenticationTypes

from polyaxon_client import settings
from polyaxon_client.exceptions import PolyaxonClientException


class ApiConfig(object):
    PAGE_SIZE = 20
    BASE_URL = "{}/api/{}"
    BASE_WS_URL = "{}/ws/{}"

    def __init__(self,
                 host=None,
                 http_port=None,
                 ws_port=None,
                 token=None,
                 version=None,
                 authentication_type=None,
                 use_https=None,
                 verify_ssl=None,
                 in_cluster=None,
                 schema_response=None,
                 reraise=False,
                 timeout=None,
                 interval=None):

        self.token = token or settings.SECRET_USER_TOKEN
        self.host = host or settings.API_HOST
        self.in_cluster = self._get_bool(in_cluster, settings.IN_CLUSTER)
        self.use_https = self._get_bool(use_https, settings.USE_HTTPS)
        self.verify_ssl = self._get_bool(verify_ssl, settings.VERIFY_SSL)

        if not self.host and not self.in_cluster:
            raise PolyaxonClientException(
                'Api config requires at least a host if not running in-cluster.')

        self.http_port = http_port or settings.HTTP_PORT or (settings.DEFAULT_HTTPS_PORT
                                                             if self.use_https
                                                             else settings.DEFAULT_HTTP_PORT)
        self.ws_port = ws_port or settings.WS_PORT or (settings.DEFAULT_HTTPS_PORT
                                                       if self.use_https
                                                       else settings.DEFAULT_HTTP_PORT)
        self.version = version or settings.API_VERSION
        self.internal_header = None

        if self.in_cluster:
            if not settings.API_HTTP_HOST:
                print('Could get api host info, '
                      'please make sure this is running inside a polyaxon job.')
            self.http_host = settings.API_HTTP_HOST
            self.ws_host = settings.API_WS_HOST
            if all([settings.INTERNAL_HEADER, settings.INTERNAL_HEADER_SERVICE]):
                self.internal_header = {settings.INTERNAL_HEADER: settings.INTERNAL_HEADER_SERVICE}

            internal_token_cond = (
                self.internal_header and
                not self.token and
                settings.SECRET_INTERNAL_TOKEN
            )
            if internal_token_cond:
                self.token = settings.SECRET_INTERNAL_TOKEN
        else:
            http_protocol = 'https' if self.use_https else 'http'
            ws_protocol = 'wss' if self.use_https else 'ws'
            self.http_host = '{}://{}:{}'.format(http_protocol, self.host, self.http_port)
            self.ws_host = '{}://{}:{}'.format(ws_protocol, self.host, self.ws_port)
        self.base_url = self.BASE_URL.format(self.http_host, self.version)
        self.base_ws_url = self.BASE_WS_URL.format(self.ws_host, self.version)
        self.authentication_type = (authentication_type or
                                    settings.AUTHENTICATION_TYPE or
                                    AuthenticationTypes.TOKEN)
        self.schema_response = self._get_bool(schema_response, settings.SCHEMA_RESPONSE)
        self.reraise = reraise
        self.timeout = timeout if timeout is not None else settings.TIMEOUT
        self.interval = interval if timeout is not None else settings.INTERVAL

    @staticmethod
    def _get_bool(value, default_value):
        if isinstance(value, bool):
            return value

        return default_value



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from polyaxon_client.api.base import BaseApiHandler
from polyaxon_client.exceptions import PolyaxonClientException
from polyaxon_client.schemas import JobConfig, JobStatusConfig


class JobApi(BaseApiHandler):
    """
    Api handler to get jobs from the server.
    """
    ENDPOINT = "/"

    def get_job(self, username, project_name, job_id):
        request_url = self.build_url(self._get_http_url(),
                                     username,
                                     project_name,
                                     'jobs',
                                     job_id)
        try:
            response = self.transport.get(request_url)
            return self.prepare_results(response_json=response.json(), config=JobConfig)
        except PolyaxonClientException as e:
            self.transport.handle_exception(e=e, log_message='Error while retrieving job.')
            return None

    def update_job(self, username, project_name, job_id, patch_dict, background=False):
        request_url = self.build_url(self._get_http_url(),
                                     username,
                                     project_name,
                                     'jobs',
                                     job_id)

        if background:
            self.transport.async_patch(request_url, json_data=patch_dict)
            return None

        try:
            response = self.transport.patch(request_url, json_data=patch_dict)
            return self.prepare_results(response_json=response.json(), config=JobConfig)
        except PolyaxonClientException as e:
            self.transport.handle_exception(e=e, log_message='Error while updating job.')
            return None

    def delete_job(self, username, project_name, job_id, background=False):
        request_url = self.build_url(self._get_http_url(),
                                     username,
                                     project_name,
                                     'jobs',
                                     job_id)

        if background:
            self.transport.async_delete(request_url)
            return None

        try:
            return self.transport.delete(request_url)
        except PolyaxonClientException as e:
            self.transport.handle_exception(e=e, log_message='Error while deleting job.')
            return None

    def get_statuses(self, username, project_name, job_id, page=1):
        request_url = self.build_url(self._get_http_url(),
                                     username,
                                     project_name,
                                     'jobs',
                                     job_id,
                                     'statuses')
        try:
            response = self.transport.get(request_url, params=self.get_page(page=page))
            return self.prepare_list_results(response.json(), page, JobStatusConfig)
        except PolyaxonClientException as e:
            self.transport.handle_exception(e=e, log_message='Error while retrieving job statuses.')
            return None

    def create_status(self,
                      username,
                      project_name,
                      job_id,
                      status,
                      message=None,
                      traceback=None,
                      background=False):
        request_url = self.build_url(self._get_http_url(),
                                     username,
                                     project_name,
                                     'jobs',
                                     job_id,
                                     'statuses')

        json_data = {'status': status}
        if message:
            json_data['message'] = message
        if traceback:
            json_data['traceback'] = traceback
        if background:
            self.transport.async_post(request_url, json_data=json_data)
            return None

        try:
            response = self.transport.post(request_url, json_data=json_data)
            return self.prepare_results(response_json=response.json(),
                                        config=JobStatusConfig)
        except PolyaxonClientException as e:
            self.transport.handle_exception(
                e=e, log_message='Error while creating job status.')
            return None

    def restart(self,
                username,
                project_name,
                job_id,
                config=None,
                update_code=None,
                background=False):
        """Restart an job."""
        request_url = self.build_url(self._get_http_url(),
                                     username,
                                     project_name,
                                     'jobs',
                                     job_id,
                                     'restart')

        data = {}
        if config:
            data['config'] = config
        if update_code:
            data['update_code'] = update_code

        if background:
            self.transport.async_post(request_url, json_data=data)
            return None

        try:
            response = self.transport.post(request_url, json_data=data)
            return self.prepare_results(response_json=response.json(), config=JobConfig)
        except PolyaxonClientException as e:
            self.transport.handle_exception(e=e, log_message='Error while restarting the job.')
            return None

    def resume(self,
               username,
               project_name,
               job_id,
               config=None,
               update_code=None,
               background=False):
        """Resume a job."""
        request_url = self.build_url(self._get_http_url(),
                                     username,
                                     project_name,
                                     'jobs',
                                     job_id,
                                     'resume')

        data = {}
        if config:
            data['config'] = config
        if update_code:
            data['update_code'] = update_code

        if background:
            self.transport.async_post(request_url, json_data=data)
            return None

        try:
            response = self.transport.post(request_url, json_data=data)
            return self.prepare_results(response_json=response.json(), config=JobConfig)
        except PolyaxonClientException as e:
            self.transport.handle_exception(e=e, log_message='Error while resuming the job.')
            return None

    def copy(self,
             username,
             project_name,
             job_id,
             config=None,
             update_code=None,
             background=False):
        """Copy an job."""
        request_url = self.build_url(self._get_http_url(),
                                     username,
                                     project_name,
                                     'jobs',
                                     job_id,
                                     'copy')

        data = {}
        if config:
            data['config'] = config
        if update_code:
            data['update_code'] = update_code

        if background:
            self.transport.async_post(request_url, json_data=data)
            return None

        try:
            response = self.transport.post(request_url, json_data=data)
            return self.prepare_results(response_json=response.json(), config=JobConfig)
        except PolyaxonClientException as e:
            self.transport.handle_exception(e=e, log_message='Error while copying the job.')
            return None

    def stop(self, username, project_name, job_id, background=False):
        request_url = self.build_url(self._get_http_url(),
                                     username,
                                     project_name,
                                     'jobs',
                                     job_id,
                                     'stop')

        if background:
            self.transport.async_post(request_url)
            return None

        try:
            return self.transport.post(request_url)
        except PolyaxonClientException as e:
            self.transport.handle_exception(e=e, log_message='Error while stopping job.')
            return None

    def bookmark(self, username, project_name, job_id, background=False):
        request_url = self.build_url(self._get_http_url(),
                                     username,
                                     project_name,
                                     'jobs',
                                     job_id,
                                     'bookmark')

        if background:
            self.transport.async_post(request_url)
            return None

        try:
            return self.transport.post(request_url)
        except PolyaxonClientException as e:
            self.transport.handle_exception(e=e, log_message='Error while bookmarking job.')
            return None

    def unbookmark(self, username, project_name, job_id, background=False):
        request_url = self.build_url(self._get_http_url(),
                                     username,
                                     project_name,
                                     'jobs',
                                     job_id,
                                     'unbookmark')

        if background:
            self.transport.async_delete(request_url)
            return None

        try:
            return self.transport.delete(request_url)
        except PolyaxonClientException as e:
            self.transport.handle_exception(e=e, log_message='Error while unbookmarking job.')
            return None

    def resources(self, username, project_name, job_id, message_handler=None):
        """Streams jobs resources using websockets.

        message_handler: handles the messages received from server.
            e.g. def f(x): print(x)
        """
        request_url = self.build_url(self._get_ws_url(),
                                     username,
                                     project_name,
                                     'jobs',
                                     job_id,
                                     'resources')
        self.transport.stream(request_url, message_handler=message_handler)

    # pylint:disable=inconsistent-return-statements
    def logs(self, username, project_name, job_id, stream=True, message_handler=None):
        """Streams jobs logs using websockets.

        message_handler: handles the messages received from server.
            e.g. def f(x): print(x)
        """
        if not stream:
            request_url = self.build_url(self._get_http_url(),
                                         username,
                                         project_name,
                                         'jobs',
                                         job_id,
                                         'logs')

            try:
                return self.transport.get(request_url)
            except PolyaxonClientException as e:
                self.transport.handle_exception(e=e, log_message='Error while retrieving jobs.')
                return []

        request_url = self.build_url(self._get_ws_url(),
                                     username,
                                     project_name,
                                     'jobs',
                                     job_id,
                                     'logs')
        self.transport.stream(request_url, message_handler=message_handler)

    def download_outputs(self, username, project_name, job_id):
        """Downloads outputs for this job to the current dir."""
        request_url = self.build_url(self._get_http_url(),
                                     username,
                                     project_name,
                                     'jobs',
                                     job_id,
                                     'outputs',
                                     'download')

        try:
            response = self.transport.download(
                request_url,
                '{}.{}.{}.tar.gz'.format(username, project_name, job_id))
            return response
        except PolyaxonClientException as e:
            self.transport.handle_exception(e=e, log_message='Error while downloading job outputs.')
            return None

    def get_heartbeat_url(self, username, project_name, job_id):
        return self.build_url(self._get_http_url(),
                              username,
                              project_name,
                              'jobs',
                              job_id,
                              self.HEARTBEAT)



# -*- coding: utf-8 -*-

from polyaxon_client.tracking.build import BuildJob
from polyaxon_client.tracking.experiment import Experiment
from polyaxon_client.tracking.group import Group
from polyaxon_client.tracking.job import Job
from polyaxon_client.tracking.paths import *



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

import json
import os

from unittest import TestCase


class TestEnvVarsCase(TestCase):
    @staticmethod
    def check_empty_value(key, expected_function):
        os.environ.pop(key, None)
        assert expected_function() is None

    @staticmethod
    def check_non_dict_value(key, expected_function, value='non dict random value'):
        os.environ[key] = value
        assert expected_function() is None

    @staticmethod
    def check_valid_dict_value(key, expected_function, value):
        os.environ[key] = json.dumps(value)
        assert expected_function() == value

    @staticmethod
    def check_valid_value(key, expected_function, value):
        os.environ[key] = value
        assert expected_function() == value



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

import httpretty
import json

from tests.test_api.utils import TestBaseApi

from polyaxon_client.api.base import BaseApiHandler
from polyaxon_client.api.version import VersionApi
from polyaxon_client.schemas import (
    ChartVersionConfig,
    CliVersionConfig,
    LibVersionConfig,
    LogHandlerConfig,
    PlatformVersionConfig
)


class TestVersionApi(TestBaseApi):

    def setUp(self):
        super(TestVersionApi, self).setUp()
        self.api_handler = VersionApi(transport=self.transport, config=self.api_config)

    @httpretty.activate
    def test_get_cli_version(self):
        obj = CliVersionConfig(latest_version='1.0', min_version='0.5').to_dict()
        httpretty.register_uri(
            httpretty.GET,
            BaseApiHandler.build_url(
                self.api_config.base_url,
                '/versions/',
                'cli'),
            body=json.dumps(obj),
            content_type='application/json', status=200)

        # Schema response
        result = self.api_handler.get_cli_version()
        assert result.to_dict() == obj

        # Raw response
        self.set_raw_response()
        result = self.api_handler.get_cli_version()
        assert result == obj

    @httpretty.activate
    def test_get_platform_version(self):
        obj = PlatformVersionConfig(latest_version='1.0', min_version='0.5').to_dict()
        httpretty.register_uri(
            httpretty.GET,
            BaseApiHandler.build_url(
                self.api_config.base_url,
                '/versions/',
                'platform'),
            body=json.dumps(obj),
            content_type='application/json', status=200)

        # Schema response
        result = self.api_handler.get_platform_version()
        assert result.to_dict() == obj

        # Raw response
        self.set_raw_response()
        result = self.api_handler.get_platform_version()
        assert result == obj

    @httpretty.activate
    def test_get_lib_version(self):
        obj = LibVersionConfig(latest_version='1.0', min_version='0.5').to_dict()
        httpretty.register_uri(
            httpretty.GET,
            BaseApiHandler.build_url(
                self.api_config.base_url,
                '/versions/',
                'lib'),
            body=json.dumps(obj),
            content_type='application/json', status=200)

        # Schema response
        result = self.api_handler.get_lib_version()
        assert result.to_dict() == obj

        # Raw response
        self.set_raw_response()
        result = self.api_handler.get_lib_version()
        assert result == obj

    @httpretty.activate
    def test_get_chart_version(self):
        obj = ChartVersionConfig(version='1.0').to_dict()
        httpretty.register_uri(
            httpretty.GET,
            BaseApiHandler.build_url(
                self.api_config.base_url,
                '/versions/',
                'chart'),
            body=json.dumps(obj),
            content_type='application/json', status=200)

        # Schema response
        result = self.api_handler.get_chart_version()
        assert result.to_dict() == obj

        # Raw response
        self.set_raw_response()
        result = self.api_handler.get_chart_version()
        assert result == obj

    @httpretty.activate
    def test_get_log_handler(self):
        obj = LogHandlerConfig(
            dsn='test',
            environment='staging',
            tags={
                'cli_min_version': '0.0.1',
                'cli_latest_version': '0.0.2',
                'platform_min_version': '0.0.1',
                'platform_latest_version': '0.0.1',
                'chart_version': '0.0.1'
            }
        ).to_dict()
        httpretty.register_uri(
            httpretty.GET,
            BaseApiHandler.build_url(
                self.api_config.base_url,
                '/log_handler'),
            body=json.dumps(obj),
            content_type='application/json', status=200)

        # Schema response
        result = self.api_handler.get_log_handler()
        assert result.to_dict() == obj

        # Raw response
        self.set_raw_response()
        result = self.api_handler.get_log_handler()
        assert result == obj



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

import six

from collections import Mapping

from polyaxon_client import settings
from polyaxon_client.exceptions import ERRORS_MAPPING, PolyaxonClientException


class BaseApiHandler(object):
    """
    Base api handler.
    """
    ENDPOINT = None
    HEARTBEAT = '_heartbeat'

    def __init__(self, transport, config):
        self.transport = transport
        self.config = config

    @staticmethod
    def build_url(*parts):
        url = ''
        for part in parts:
            part = part.rstrip('/').lstrip('/') if isinstance(part, six.string_types) else part
            if part:
                url += '{}/'.format(part)

        return url

    def _get_url(self, base_url, endpoint=None):
        endpoint = endpoint or self.ENDPOINT
        if endpoint is None:
            raise ERRORS_MAPPING['base'](
                "This function expects `ENDPOINT` attribute to be set, "
                "or an `endpoint` argument to be passed.")
        return self.build_url(base_url, endpoint)

    def _get_http_url(self, endpoint=None):
        return self._get_url(self.config.base_url, endpoint)

    def _get_ws_url(self, endpoint=None):
        return self._get_url(self.config.base_ws_url, endpoint)

    def get_page(self, page=1):
        if page <= 1:
            return {}
        return {'offset': (page - 1) * self.config.PAGE_SIZE}

    def prepare_list_results(self, response_json, current_page, config):
        list_results = {
            'count': response_json.get('count', 0),
            'next': current_page + 1 if response_json.get('next') else None,
            'previous': current_page - 1 if response_json.get('previous') else None,
        }
        if self.config.schema_response:
            list_results['results'] = [
                config.from_dict(obj, unknown=settings.RECEPTION_UNKNOWN_BEHAVIOUR)
                for obj in response_json.get('results', [])]
        else:
            list_results['results'] = response_json.get('results', [])

        return list_results

    def prepare_results(self, response_json, config):
        return (config.from_dict(response_json, unknown=settings.RECEPTION_UNKNOWN_BEHAVIOUR)
                if self.config.schema_response else response_json)

    @staticmethod
    def validate_config(config, config_schema):
        if isinstance(config, Mapping):
            return config_schema.from_dict(config, unknown=settings.VALIDATION_UNKNOWN_BEHAVIOUR)
        elif not isinstance(config, config_schema):
            raise PolyaxonClientException(
                'Received an invalid config. '
                'Expects a Mapping or an instance of `{}`.'.format(config_schema.__name__))

        return config



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from polyaxon_client.exceptions import PolyaxonClientException
from polyaxon_client.logger import logger
from polyaxon_client.transport.retry_transport import RetryTransportMixin
from polyaxon_client.workers.periodic_worker import PeriodicWorker


class PeriodicHttpTransportMixin(RetryTransportMixin):
    """Periodic http operations transport."""

    @property
    def periodic_http_done(self):
        if hasattr(self, '_periodic_http_done'):
            return self._periodic_http_done
        return None

    @property
    def periodic_http_exceptions(self):
        if hasattr(self, '_periodic_http_exceptions'):
            return self._periodic_http_exceptions
        return None

    def queue_periodic_request(self, request, url, **kwargs):
        try:
            request(url=url, session=self.retry_session, **kwargs)
        except Exception as e:
            self._periodic_http_exceptions += 1
            logger.debug('Error making request url: %s, params: params %s, exp: %s', url, kwargs, e)
        finally:
            self._periodic_http_done += 1

    @property
    def periodic_http_worker(self):
        if hasattr(self, '_periodic_http_worker'):
            return self._periodic_http_worker
        return None

    def get_periodic_http_worker(self, **kwargs):
        worker = self.periodic_http_worker
        if not worker or not worker.is_alive():
            if 'request' not in kwargs:
                raise PolyaxonClientException('Periodic worker expects a request argument.')
            self._periodic_http_worker = PeriodicWorker(
                callback=self.queue_periodic_request,
                worker_interval=self.config.interval,
                worker_timeout=self.config.timeout,
                kwargs=kwargs)
            self._periodic_http_worker.start()
        return self.periodic_http_worker

    def set_health_check(self, url):
        worker = self.get_periodic_http_worker(request=self.post)
        worker.queue_health(url)

    def periodic_post(self,
                      url,
                      params=None,
                      data=None,
                      files=None,
                      json_data=None,
                      timeout=None,
                      headers=None):
        """Periodic Async Call request with a post."""
        worker = self.get_periodic_http_worker(request=self.post,
                                               params=params,
                                               files=files,
                                               timeout=timeout,
                                               headers=headers)
        return worker.queue(url=url, data=data, json_data=json_data)


class PeriodicWSTransportMixin(RetryTransportMixin):
    """Periodic websocket operations transport."""

    @property
    def periodic_ws_done(self):
        if hasattr(self, '_periodic_ws_done'):
            return self._periodic_ws_done
        return None

    @property
    def periodic_ws_exceptions(self):
        if hasattr(self, '_periodic_ws_exceptions'):
            return self._periodic_ws_exceptions
        return None

    def queue_ws_request(self, request, url, **kwargs):
        try:
            request(url=url, **kwargs)
        except Exception as e:
            self._periodic_ws_exceptions += 1
            logger.debug('Error making request url: %s, params: params %s, exp: %s', url, kwargs, e)
        finally:
            self._periodic_ws_done += 1

    @property
    def periodic_ws_workers(self):
        if not hasattr(self, '_periodic_ws_workers'):
            self._periodic_ws_workers = {}

        return self._periodic_ws_workers

    def get_periodic_http_worker(self, url, **kwargs):
        worker = self.periodic_ws_workers.get(url)
        if not worker or not worker.is_alive():
            kwargs['url'] = url
            kwargs['request'] = self.socket(url, message_handler=None, **kwargs)
            worker = PeriodicWorker(callback=self.queue_ws_request,
                                    worker_interval=self.config.interval,
                                    worker_timeout=self.config.timeout,
                                    kwargs=kwargs)
            worker.start()
            self.periodic_ws_workers[url] = worker
        return self.periodic_ws_workers[url]

    def periodic_send(self, url, data=None, headers=None):
        """Periodic Async Call request with a post."""
        worker = self.get_periodic_http_worker(url=url, headers=headers)
        return worker.queue(data=data)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

import json
import os
import requests
import tarfile

from clint.textui import progress
from clint.textui.progress import Bar
from hestia.list_utils import to_list
from requests_toolbelt import MultipartEncoder, MultipartEncoderMonitor

from polyaxon_client import settings
from polyaxon_client.exceptions import ERRORS_MAPPING, PolyaxonShouldExitError
from polyaxon_client.logger import logger


class HttpTransportMixin(object):
    """HTTP operations transport."""
    MAX_UPLOAD_SIZE = 1024 * 1024 * 150

    @property
    def session(self):
        if not hasattr(self, '_session'):
            self._session = requests.Session()
        return self._session

    @staticmethod
    def create_progress_callback(encoder):
        encoder_len = encoder.len
        progress_bar = Bar(expected_size=encoder_len, filled_char='=')

        def callback(monitor):
            progress_bar.show(monitor.bytes_read)

        return callback, progress_bar

    @staticmethod
    def format_sizeof(num, suffix='B'):
        """
        Print in human friendly format
        """
        for unit in ['', 'Ki', 'Mi', 'Gi', 'Ti', 'Pi', 'Ei', 'Zi']:
            if abs(num) < 1024.0:
                return "%3.1f%s%s" % (num, unit, suffix)
            num /= 1024.0
        return "%.1f%s%s" % (num, 'Yi', suffix)

    def request(self,
                method,
                url,
                params=None,
                data=None,
                files=None,
                json=None,  # noqa
                timeout=None,
                headers=None,
                session=None):
        """Send a request with the given data as json to the given URL.

        Args:
            method: HTTP method
            url: The url to send the request to.
            params: Dictionary of values to format url.
            data:
            files:
            json:
            timeout:
            headers: extra headers to add to the request.

        Returns:
            Request response if successful, Exception otherwise.

        Raises:
            PolyaxonHTTPError or one of its subclasses.
        """
        logger.debug("Starting request to url: %s with params: %s, data: %s",
                     url, params, data)

        request_headers = self._get_headers(headers=headers)
        timeout = timeout if timeout is not None else settings.LONG_REQUEST_TIMEOUT
        session = session or self.session

        try:
            response = session.request(method,
                                       url,
                                       params=params,
                                       data=data,
                                       json=json,
                                       headers=request_headers,
                                       files=files,
                                       timeout=timeout,
                                       verify=self.config.verify_ssl)
        except (requests.exceptions.RequestException,
                requests.exceptions.Timeout,
                requests.exceptions.HTTPError) as exception:
            try:
                logger.debug("Exception: %s", exception, exc_info=True)
            except TypeError:
                pass
            raise PolyaxonShouldExitError(
                "Error connecting to Polyaxon server on `{}`.\n"
                "An Error `{}` occurred.\n"
                "Check your host and ports configuration "
                "and your internet connection.".format(url, exception))

        logger.debug("Response Content: %s, Headers: %s", response.content, response.headers)
        self.check_response_status(response, url)
        return response

    def upload(self,
               url,
               files,
               files_size,
               params=None,
               json_data=None,
               timeout=None,
               headers=None,
               session=None):

        if files_size > self.MAX_UPLOAD_SIZE:
            raise PolyaxonShouldExitError(
                "Files too large to sync, please keep it under {}.\n"
                "If you have data files in the current directory, "
                "please add them directly to your data volume, or upload them "
                "separately using `polyaxon data` command and remove them from here.\n".format(
                    self.format_sizeof(self.MAX_UPLOAD_SIZE)))

        files = to_list(files)
        if json_data:
            files.append(('json', json.dumps(json_data)))

        multipart_encoder = MultipartEncoder(
            fields=files
        )
        request_headers = headers or {}
        request_headers.update({"Content-Type": multipart_encoder.content_type})

        # Attach progress bar
        progress_callback, progress_bar = self.create_progress_callback(multipart_encoder)
        multipart_encoder_monitor = MultipartEncoderMonitor(multipart_encoder, progress_callback)

        timeout = timeout if timeout is not None else settings.LONG_REQUEST_TIMEOUT

        try:
            response = self.put(url=url,
                                params=params,
                                data=multipart_encoder_monitor,
                                headers=request_headers,
                                timeout=timeout,
                                session=session)
        finally:
            # always make sure we clear the console
            progress_bar.done()

        return response

    def download(self,
                 url,
                 filename,
                 params=None,
                 headers=None,
                 timeout=None,
                 session=None,
                 untar=False,
                 delete_tar=True,
                 extract_path=None):
        """
        Download the file from the given url at the current path
        """
        logger.debug("Downloading files from url: %s", url)

        request_headers = self._get_headers(headers=headers)
        timeout = timeout if timeout is not None else settings.LONG_REQUEST_TIMEOUT
        session = session or self.session

        try:
            response = session.get(url,
                                   params=params,
                                   headers=request_headers,
                                   timeout=timeout,
                                   stream=True)
            self.check_response_status(response, url)
            with open(filename, 'wb') as f:
                # chunk mode response doesn't have content-length so we are
                # using a custom header here
                content_length = response.headers.get('x-polyaxon-content-length')
                if not content_length:
                    content_length = response.headers.get('content-length')
                if content_length:
                    for chunk in progress.bar(response.iter_content(chunk_size=1024),
                                              expected_size=(int(content_length) / 1024) + 1):
                        if chunk:
                            f.write(chunk)
                else:
                    for chunk in response.iter_content(chunk_size=1024):
                        if chunk:
                            f.write(chunk)

            if untar:
                filename = self.untar(filename=filename,
                                      delete_tar=delete_tar,
                                      extract_path=extract_path)
            return filename
        except (requests.exceptions.ConnectionError,
                requests.exceptions.RequestException,
                requests.exceptions.Timeout,
                requests.exceptions.HTTPError) as exception:
            try:
                logger.debug("Exception: %s", exception)
            except TypeError:
                pass

            raise PolyaxonShouldExitError(
                "Error connecting to Polyaxon server on `{}`.\n"
                "An Error `{}` occurred.\n"
                "Check your host and ports configuration "
                "and your internet connection.".format(url, exception))

    def untar(self, filename, delete_tar=True, extract_path=None):
        extract_path = extract_path or '.'
        logger.info("Untarring the contents of the file ...")
        tar = tarfile.open(filename)
        tar.extractall(extract_path)
        tar.close()
        if delete_tar:
            logger.info("Cleaning up the tar file ...")
            os.remove(filename)
        return filename

    def check_response_status(self, response, endpoint):
        """Check if response is successful. Else raise Exception."""

        if 200 <= response.status_code < 300:
            return response

        try:
            logger.error(
                "Request to %s failed with status code %s. \n"
                "Reason: %s", endpoint, response.status_code, response.text)
        except TypeError:
            logger.error("Request to %s failed with status code", endpoint)

        exception = ERRORS_MAPPING.get(response.status_code, ERRORS_MAPPING['http'])
        raise exception(endpoint=endpoint,
                        response=response,
                        message=response.text,
                        status_code=response.status_code)

    def get(self,
            url,
            params=None,
            data=None,
            files=None,
            json_data=None,
            timeout=None,
            headers=None,
            session=None):
        """Call request with a get."""
        return self.request('GET',
                            url=url,
                            params=params,
                            data=data,
                            files=files,
                            json=json_data,
                            timeout=timeout,
                            headers=headers,
                            session=session)

    def post(self,
             url,
             params=None,
             data=None,
             files=None,
             json_data=None,
             timeout=None,
             headers=None,
             session=None):
        """Call request with a post."""
        return self.request('POST',
                            url=url,
                            params=params,
                            data=data,
                            files=files,
                            json=json_data,
                            timeout=timeout,
                            headers=headers,
                            session=session)

    def patch(self,
              url,
              params=None,
              data=None,
              files=None,
              json_data=None,
              timeout=None,
              headers=None,
              session=None):
        """Call request with a patch."""
        return self.request('PATCH',
                            url=url,
                            params=params,
                            data=data,
                            files=files,
                            json=json_data,
                            timeout=timeout,
                            headers=headers,
                            session=session)

    def delete(self,
               url,
               params=None,
               data=None,
               files=None,
               json_data=None,
               timeout=None,
               headers=None,
               session=None):
        """Call request with a delete."""
        return self.request('DELETE',
                            url=url,
                            params=params,
                            data=data,
                            files=files,
                            json=json_data,
                            timeout=timeout,
                            headers=headers,
                            session=session)

    def put(self,
            url,
            params=None,
            data=None,
            files=None,
            json_data=None,
            timeout=None,
            headers=None,
            session=None):
        """Call request with a put."""
        return self.request('PUT',
                            url=url,
                            params=params,
                            data=data,
                            files=files,
                            json=json_data,
                            timeout=timeout,
                            headers=headers,
                            session=session)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

import tempfile
import time

from unittest import TestCase

from mock import patch

from polyaxon_client import settings
from polyaxon_client.api_config import ApiConfig
from polyaxon_client.transport import Transport


class TestBaseApi(TestCase):
    def setUp(self):
        self.host = 'localhost'
        self.http_port = 8000
        self.ws_port = 1337
        self.api_config = ApiConfig(host=self.host,
                                    http_port=self.http_port,
                                    ws_port=self.ws_port,
                                    version='v1',
                                    token='token',
                                    reraise=True,
                                    use_https=False,
                                    in_cluster=False,
                                    interval=0,
                                    timeout=0.01,
                                    schema_response=True)
        self.transport = Transport(config=self.api_config)
        settings.TMP_AUTH_TOKEN_PATH = '{}/{}'.format(tempfile.mkdtemp(), '.plx')
        settings.CLIENT_CONFIG_PATH = '{}/{}'.format(tempfile.mkdtemp(), '.plx')

    def set_raw_response(self):
        self.api_config.schema_response = False

    def set_schema_response(self):
        self.api_config.schema_response = True

    def assert_async_call(self, api_handler_call, method):
        with patch.object(Transport, method) as mock_fct:
            api_handler_call()
        time.sleep(0.02)
        assert mock_fct.call_count == 1

    def assert_periodic_call(self, api_handler_call, method):
        with patch.object(Transport, method) as mock_fct:
            api_handler_call()
        time.sleep(0.01)
        assert mock_fct.call_count == 1



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

import httpretty
import json
import os
import tempfile
import uuid

from tests.test_api.utils import TestBaseApi

from polyaxon_client import settings
from polyaxon_client.api.auth import AuthApi
from polyaxon_client.api.base import BaseApiHandler
from polyaxon_client.schemas import CredentialsConfig, UserConfig


class TestAuthApi(TestBaseApi):

    def setUp(self):
        super(TestAuthApi, self).setUp()
        settings.CONTEXT_AUTH_TOKEN_PATH = '{}/{}'.format(tempfile.mkdtemp(), '.authtoken')
        settings.TMP_AUTH_TOKEN_PATH = '{}/{}'.format(tempfile.mkdtemp(), '.authtoken')
        self.api_handler = AuthApi(transport=self.transport, config=self.api_config)

    @httpretty.activate
    def test_get_user(self):
        user = UserConfig('user', 'user@test.com').to_dict()
        httpretty.register_uri(
            httpretty.GET,
            BaseApiHandler.build_url(
                self.api_config.base_url,
                '/users'),
            body=json.dumps(user),
            content_type='application/json', status=200)

        # Schema response
        result = self.api_handler.get_user('token_value')
        assert result.to_dict() == user

        # Raw response
        self.set_raw_response()
        result = self.api_handler.get_user('token_value')
        assert result == user

    @httpretty.activate
    def test_login(self):
        token = uuid.uuid4().hex
        httpretty.register_uri(
            httpretty.POST,
            BaseApiHandler.build_url(
                self.api_config.base_url,
                '/users',
                'token'
            ),
            body=json.dumps({'token': token}),
            content_type='application/json', status=200)

        # Login without updating the token
        assert self.api_config.token == 'token'
        credentials = CredentialsConfig('user', 'password')
        assert token == self.api_handler.login(credentials=credentials, set_token=False)
        assert self.api_config.token == 'token'

        # Login and update the token
        assert self.api_config.token == 'token'
        assert token == self.api_handler.login(credentials=credentials, set_token=True)
        assert self.api_config.token == token

    @httpretty.activate
    def test_login_experiment_ephemeral_token(self):
        token = uuid.uuid4().hex
        httpretty.register_uri(
            httpretty.POST,
            BaseApiHandler.build_url(
                self.api_config.base_url,
                '/',
                'user',
                'project',
                'experiments',
                '1',
                'ephemeraltoken'
            ),
            body=json.dumps({'token': token}),
            content_type='application/json', status=200)

        # Login without updating the token and without persistence
        if os.path.exists(settings.TMP_AUTH_TOKEN_PATH):
            os.remove(settings.TMP_AUTH_TOKEN_PATH)
        assert self.api_config.token == 'token'
        assert token == self.api_handler.login_experiment_ephemeral_token(
            username='user',
            project_name='project',
            experiment_id=1,
            ephemeral_token='foo',
            set_token=False,
            persist_token=False)
        assert self.api_config.token == 'token'
        assert os.path.exists(settings.TMP_AUTH_TOKEN_PATH) is False

        # Login and update the token and persistence
        if os.path.exists(settings.TMP_AUTH_TOKEN_PATH):
            os.remove(settings.TMP_AUTH_TOKEN_PATH)
        assert self.api_config.token == 'token'
        assert token == self.api_handler.login_experiment_ephemeral_token(
            username='user',
            project_name='project',
            experiment_id=1,
            ephemeral_token='foo',
            set_token=True,
            persist_token=True)
        assert self.api_config.token == token
        assert os.path.exists(settings.TMP_AUTH_TOKEN_PATH) is True

        # Login remove ephemeral token from env var and settings
        os.environ[settings.SECRET_EPHEMERAL_TOKEN_KEY] = 'value'
        settings.SECRET_EPHEMERAL_TOKEN = 'eph_token'  # noqa
        if os.path.exists(settings.TMP_AUTH_TOKEN_PATH):
            os.remove(settings.TMP_AUTH_TOKEN_PATH)
        assert self.api_config.token == token
        assert os.environ.get(settings.SECRET_EPHEMERAL_TOKEN_KEY) == 'value'
        assert settings.SECRET_EPHEMERAL_TOKEN == 'eph_token'
        assert token == self.api_handler.login_experiment_ephemeral_token(
            username='user',
            project_name='project',
            experiment_id=1,
            ephemeral_token='foo',
            set_token=True,
            persist_token=True)
        assert self.api_config.token == token
        assert os.path.exists(settings.TMP_AUTH_TOKEN_PATH) is True
        assert os.environ.get(settings.SECRET_EPHEMERAL_TOKEN_KEY) is None
        assert not hasattr(settings, 'SECRET_EPHEMERAL_TOKEN')

        assert token == self.api_handler.login_experiment_ephemeral_token(
            username='user',
            project_name='project',
            experiment_id=1,
            ephemeral_token='foo',
            set_token=True,
            persist_token=True)

    @httpretty.activate
    def test_login_experiment_impersonate_token(self):
        token = uuid.uuid4().hex
        httpretty.register_uri(
            httpretty.POST,
            BaseApiHandler.build_url(
                self.api_config.base_url,
                '/',
                'user',
                'project',
                'experiments',
                '1',
                'imporsonatetoken'
            ),
            body=json.dumps({'token': token}),
            content_type='application/json', status=200)

        # Login without updating the token and without persistence
        if os.path.exists(settings.CONTEXT_AUTH_TOKEN_PATH):
            os.remove(settings.CONTEXT_AUTH_TOKEN_PATH)
        assert self.api_config.token == 'token'
        assert token == self.api_handler.login_experiment_impersonate_token(
            username='user',
            project_name='project',
            experiment_id=1,
            internal_token='foo',
            set_token=False,
            persist_token=False)
        assert self.api_config.token == 'token'
        assert os.path.exists(settings.CONTEXT_AUTH_TOKEN_PATH) is False

        # Login and update the token and persistence
        if os.path.exists(settings.CONTEXT_AUTH_TOKEN_PATH):
            os.remove(settings.CONTEXT_AUTH_TOKEN_PATH)
        assert self.api_config.token == 'token'
        assert token == self.api_handler.login_experiment_impersonate_token(
            username='user',
            project_name='project',
            experiment_id=1,
            internal_token='foo',
            set_token=True,
            persist_token=True)
        assert self.api_config.token == token
        assert os.path.exists(settings.CONTEXT_AUTH_TOKEN_PATH) is True

    @httpretty.activate
    def test_login_job_impersonate_token(self):
        token = uuid.uuid4().hex
        httpretty.register_uri(
            httpretty.POST,
            BaseApiHandler.build_url(
                self.api_config.base_url,
                '/',
                'user',
                'project',
                'jobs',
                '1',
                'imporsonatetoken'
            ),
            body=json.dumps({'token': token}),
            content_type='application/json', status=200)

        # Login without updating the token and without persistence
        if os.path.exists(settings.CONTEXT_AUTH_TOKEN_PATH):
            os.remove(settings.CONTEXT_AUTH_TOKEN_PATH)
        assert self.api_config.token == 'token'
        assert token == self.api_handler.login_job_impersonate_token(
            username='user',
            project_name='project',
            job_id=1,
            internal_token='foo',
            set_token=False,
            persist_token=False)
        assert self.api_config.token == 'token'
        assert os.path.exists(settings.CONTEXT_AUTH_TOKEN_PATH) is False

        # Login and update the token and persistence
        if os.path.exists(settings.CONTEXT_AUTH_TOKEN_PATH):
            os.remove(settings.CONTEXT_AUTH_TOKEN_PATH)
        assert self.api_config.token == 'token'
        assert token == self.api_handler.login_job_impersonate_token(
            username='user',
            project_name='project',
            job_id=1,
            internal_token='foo',
            set_token=True,
            persist_token=True)
        assert self.api_config.token == token
        assert os.path.exists(settings.CONTEXT_AUTH_TOKEN_PATH) is True

    @httpretty.activate
    def test_login_notebook_impersonate_token(self):
        token = uuid.uuid4().hex
        httpretty.register_uri(
            httpretty.POST,
            BaseApiHandler.build_url(
                self.api_config.base_url,
                '/',
                'user',
                'project',
                'notebook',
                'imporsonatetoken'
            ),
            body=json.dumps({'token': token}),
            content_type='application/json', status=200)

        # Login without updating the token and without persistence
        if os.path.exists(settings.CONTEXT_AUTH_TOKEN_PATH):
            os.remove(settings.CONTEXT_AUTH_TOKEN_PATH)
        assert self.api_config.token == 'token'
        assert token == self.api_handler.login_notebook_impersonate_token(
            username='user',
            project_name='project',
            internal_token='foo',
            set_token=False,
            persist_token=False)
        assert self.api_config.token == 'token'
        assert os.path.exists(settings.CONTEXT_AUTH_TOKEN_PATH) is False

        # Login and update the token and persistence
        if os.path.exists(settings.CONTEXT_AUTH_TOKEN_PATH):
            os.remove(settings.CONTEXT_AUTH_TOKEN_PATH)
        assert self.api_config.token == 'token'
        assert token == self.api_handler.login_notebook_impersonate_token(
            username='user',
            project_name='project',
            internal_token='foo',
            set_token=True,
            persist_token=True)
        assert self.api_config.token == token
        assert os.path.exists(settings.CONTEXT_AUTH_TOKEN_PATH) is True



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

import datetime
import httpretty
import json
import uuid

from collections import Mapping

from tests.test_api.utils import TestBaseApi

from polyaxon_client.api.base import BaseApiHandler
from polyaxon_client.api.experiment_group import ExperimentGroupApi
from polyaxon_client.schemas import ExperimentConfig, ExperimentGroupConfig, GroupStatusConfig


class TestExperimentGroupApi(TestBaseApi):

    def setUp(self):
        super(TestExperimentGroupApi, self).setUp()
        self.api_handler = ExperimentGroupApi(transport=self.transport, config=self.api_config)

    @httpretty.activate
    def test_get_experiment_group(self):
        obj = ExperimentGroupConfig(content='text',
                                    uuid=uuid.uuid4().hex,
                                    project=uuid.uuid4().hex).to_dict()
        httpretty.register_uri(
            httpretty.GET,
            BaseApiHandler.build_url(
                self.api_config.base_url,
                '/',
                'username',
                'project_name',
                'groups',
                1),
            body=json.dumps(obj),
            content_type='application/json',
            status=200)

        # Schema response
        result = self.api_handler.get_experiment_group('username', 'project_name', 1)
        assert obj == result.to_dict()

        # Raw response
        self.set_raw_response()
        result = self.api_handler.get_experiment_group('username', 'project_name', 1)
        assert obj == result

    @httpretty.activate
    def test_list_experiments(self):
        group_uuid = uuid.uuid4().hex
        project_uuid = uuid.uuid4().hex
        xp_uuid = uuid.uuid4().hex
        xps = [ExperimentConfig(uuid=xp_uuid,
                                config={},
                                project=project_uuid,
                                experiment_group=group_uuid).to_dict()
               for _ in range(10)]
        httpretty.register_uri(
            httpretty.GET,
            BaseApiHandler.build_url(
                self.api_config.base_url,
                '/',
                'username',
                'project_name',
                'experiments') + '?group=1',
            body=json.dumps({'results': xps, 'count': 10, 'next': None}),
            content_type='application/json',
            status=200)

        # Schema response
        response = self.api_handler.list_experiments('username', 'project_name', 1)
        assert len(response['results']) == 10
        assert isinstance(response['results'][0], ExperimentConfig)
        assert response['count'] == 10
        assert response['next'] is None
        assert response['previous'] is None

        # Raw response
        self.set_raw_response()
        response = self.api_handler.list_experiments('username', 'project_name', 1)
        assert len(response['results']) == 10
        assert isinstance(response['results'][0], Mapping)
        assert response['count'] == 10
        assert response['next'] is None
        assert response['previous'] is None

        # Pagination
        httpretty.register_uri(
            httpretty.GET,
            BaseApiHandler.build_url(
                self.api_config.base_url,
                '/',
                'username',
                'project_name',
                'experiments') + '?group=1&offset=2',
            body=json.dumps({'results': xps, 'count': 10, 'next': None}),
            content_type='application/json',
            status=200)

        # Schema response
        self.set_schema_response()
        response = self.api_handler.list_experiments('username', 'project_name', 1, page=2)
        assert len(response['results']) == 10
        assert isinstance(response['results'][0], ExperimentConfig)

        # Raw response
        self.set_raw_response()
        response = self.api_handler.list_experiments('username', 'project_name', 1, page=2)
        assert len(response['results']) == 10
        assert isinstance(response['results'][0], Mapping)

        # Query, Sort
        httpretty.register_uri(
            httpretty.GET,
            BaseApiHandler.build_url(
                self.api_config.base_url,
                '/',
                'username',
                'project_name',
                'experiments') + '?group=1&query=started_at:>=2010-10-10,sort=created_at',
            body=json.dumps({'results': xps, 'count': 10, 'next': None}),
            content_type='application/json',
            status=200)

        # Schema response
        self.set_schema_response()
        response = self.api_handler.list_experiments('username',
                                                     'project_name',
                                                     1,
                                                     query='started_at:>=2010-10-10',
                                                     sort='created_at')
        assert len(response['results']) == 10
        assert isinstance(response['results'][0], ExperimentConfig)

        # Raw response
        self.set_raw_response()
        response = self.api_handler.list_experiments('username',
                                                     'project_name',
                                                     1,
                                                     query='started_at:>=2010-10-10',
                                                     sort='created_at')
        assert len(response['results']) == 10
        assert isinstance(response['results'][0], Mapping)

    @httpretty.activate
    def test_update_experiment_group(self):
        obj = ExperimentGroupConfig(content='text',
                                    uuid=uuid.uuid4().hex,
                                    project=uuid.uuid4().hex).to_dict()
        httpretty.register_uri(
            httpretty.PATCH,
            BaseApiHandler.build_url(
                self.api_config.base_url,
                '/',
                'username',
                'project_name',
                'groups',
                1),
            body=json.dumps(obj),
            content_type='application/json',
            status=200)

        # Schema response
        result = self.api_handler.update_experiment_group(
            'username', 'project_name', 1, {'content': 'new'})
        assert result.to_dict() == obj

        # Raw response
        self.set_raw_response()
        result = self.api_handler.update_experiment_group(
            'username', 'project_name', 1, {'content': 'new'})
        assert result == obj

        # Async
        self.assert_async_call(
            api_handler_call=lambda: self.api_handler.update_experiment_group(
                'username', 'project_name', 1, {'content': 'new'}, background=True),
            method='patch')

    @httpretty.activate
    def test_delete_experiment_group(self):
        httpretty.register_uri(
            httpretty.DELETE,
            BaseApiHandler.build_url(
                self.api_config.base_url,
                '/',
                'username',
                'project_name',
                'groups',
                1),
            content_type='application/json',
            status=204)
        result = self.api_handler.delete_experiment_group('username', 'project_name', 1)
        assert result.status_code == 204

        # Async
        self.assert_async_call(
            api_handler_call=lambda: self.api_handler.delete_experiment_group(
                'username', 'project_name', 1, background=True),
            method='delete')

    @httpretty.activate
    def test_get_experiment_group_statuses(self):
        group = GroupStatusConfig(id=1,
                                  uuid=uuid.uuid4().hex,
                                  experiment_group=1,
                                  created_at=datetime.datetime.now(),
                                  status='Running').to_dict()
        httpretty.register_uri(
            httpretty.GET,
            BaseApiHandler.build_url(
                self.api_config.base_url,
                '/',
                'username',
                'project_name',
                'groups',
                1,
                'statuses'),
            body=json.dumps({'results': [group], 'count': 1, 'next': None}),
            content_type='application/json',
            status=200)

        # Schema response
        response = self.api_handler.get_statuses('username', 'project_name', 1)
        assert len(response['results']) == 1
        assert isinstance(response['results'][0], GroupStatusConfig)

        # Raw response
        self.set_raw_response()
        response = self.api_handler.get_statuses('username', 'project_name', 1)
        assert len(response['results']) == 1
        assert isinstance(response['results'][0], Mapping)

    @httpretty.activate
    def test_create_experiment_status(self):
        exp = GroupStatusConfig(id=1,
                                uuid=uuid.uuid4().hex,
                                experiment_group=1,
                                created_at=datetime.datetime.now(),
                                status='Running').to_dict()
        httpretty.register_uri(
            httpretty.POST,
            BaseApiHandler.build_url(
                self.api_config.base_url,
                '/',
                'username',
                'project_name',
                'groups',
                1,
                'statuses'),
            body=json.dumps(exp),
            content_type='application/json',
            status=200)

        # Schema response
        response = self.api_handler.create_status('username', 'project_name', 1, status='running')
        assert response.to_dict() == exp

        # Raw response
        self.set_raw_response()
        response = self.api_handler.create_status('username', 'project_name', 1, status='running')
        assert response == exp

        # Async
        self.assert_async_call(
            api_handler_call=lambda: self.api_handler.create_status(
                'username', 'project_name', 1, status='running', background=True),
            method='post')

    @httpretty.activate
    def test_stop_experiment_group_all(self):
        httpretty.register_uri(
            httpretty.POST,
            BaseApiHandler.build_url(
                self.api_config.base_url,
                '/',
                'username',
                'project_name',
                'groups',
                1,
                'stop'),
            content_type='application/json',
            status=200)
        result = self.api_handler.stop('username', 'project_name', 1)
        assert result.status_code == 200

        # Async
        self.assert_async_call(
            api_handler_call=lambda: self.api_handler.stop(
                'username', 'project_name', 1, background=True),
            method='post')

    @httpretty.activate
    def test_stop_experiment_group_pending(self):
        httpretty.register_uri(
            httpretty.POST,
            BaseApiHandler.build_url(
                self.api_config.base_url,
                '/',
                'username',
                'project_name',
                'groups',
                1,
                'stop'),
            content_type='application/json',
            status=200)
        result = self.api_handler.stop('username', 'project_name', 1, pending=True)
        assert result.status_code == 200

        # Async
        self.assert_async_call(
            api_handler_call=lambda: self.api_handler.stop(
                'username', 'project_name', 1, pending=True, background=True),
            method='post')

    @httpretty.activate
    def test_start_experiment_group_tensorboard(self):
        httpretty.register_uri(
            httpretty.POST,
            BaseApiHandler.build_url(
                self.api_config.base_url,
                '/',
                'username',
                'project_name',
                'groups',
                1,
                'tensorboard',
                'start'),
            content_type='application/json',
            status=200)
        result = self.api_handler.start_tensorboard('username', 'project_name', 1)
        assert result.status_code == 200

        # Async
        self.assert_async_call(
            api_handler_call=lambda: self.api_handler.start_tensorboard(
                'username', 'project_name', 1, background=True),
            method='post')

    @httpretty.activate
    def test_start_experiment_group_tensorboard_with_config(self):
        obj = {}
        httpretty.register_uri(
            httpretty.POST,
            BaseApiHandler.build_url(
                self.api_config.base_url,
                '/',
                'username',
                'project_name',
                'groups',
                1,
                'tensorboard',
                'start'),
            body=json.dumps(obj),
            content_type='application/json',
            status=200)
        result = self.api_handler.start_tensorboard('username', 'project_name', 1, obj)
        assert result.status_code == 200

        # Async
        self.assert_async_call(
            api_handler_call=lambda: self.api_handler.start_tensorboard(
                'username', 'project_name', 1, obj, background=True),
            method='post')

    @httpretty.activate
    def test_stop_experiment_group_tensorboard(self):
        httpretty.register_uri(
            httpretty.POST,
            BaseApiHandler.build_url(
                self.api_config.base_url,
                '/',
                'username',
                'project_name',
                'groups',
                1,
                'tensorboard',
                'stop'),
            content_type='application/json',
            status=200)
        result = self.api_handler.stop_tensorboard('username', 'project_name', 1)
        assert result.status_code == 200

        # Async
        self.assert_async_call(
            api_handler_call=lambda: self.api_handler.stop_tensorboard(
                'username', 'project_name', 1, background=True),
            method='post')

    @httpretty.activate
    def test_bookmark_experiment(self):
        httpretty.register_uri(
            httpretty.POST,
            BaseApiHandler.build_url(
                self.api_config.base_url,
                '/',
                'username',
                'project_name',
                'groups',
                1,
                'bookmark'),
            content_type='application/json',
            status=200)
        result = self.api_handler.bookmark('username', 'project_name', 1)
        assert result.status_code == 200

        # Async
        self.assert_async_call(
            api_handler_call=lambda: self.api_handler.bookmark(
                'username', 'project_name', 1, background=True),
            method='post')

    @httpretty.activate
    def test_unbookmark_experiment(self):
        httpretty.register_uri(
            httpretty.DELETE,
            BaseApiHandler.build_url(
                self.api_config.base_url,
                '/',
                'username',
                'project_name',
                'groups',
                1,
                'unbookmark'),
            content_type='application/json',
            status=200)
        result = self.api_handler.unbookmark('username', 'project_name', 1)
        assert result.status_code == 200

        # Async
        self.assert_async_call(
            api_handler_call=lambda: self.api_handler.unbookmark(
                'username', 'project_name', 1, background=True),
            method='delete')



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

import six

from hestia.list_utils import to_list

from polyaxon_client.exceptions import PolyaxonClientException


def validate_tags(tags):
    if not tags:
        return None

    if isinstance(tags, six.string_types):
        tags = [tag.strip() for tag in tags.split(',')]

    for tag in to_list(tags):
        if not (tag and isinstance(tag, six.string_types)):
            raise PolyaxonClientException(
                'Received an invalid tag {}. '
                'Please make sure that the tags are list of strings, '
                'or a string containing comma separated values.'.format(tag))
    return tags



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

import datetime
import httpretty
import json
import uuid

from collections import Mapping

from tests.test_api.utils import TestBaseApi

from polyaxon_client.api.base import BaseApiHandler
from polyaxon_client.api.experiment_job import ExperimentJobApi
from polyaxon_client.schemas import ExperimentJobConfig, ExperimentJobStatusConfig


class TestExperimentJobClient(TestBaseApi):

    def setUp(self):
        super(TestExperimentJobClient, self).setUp()
        self.api_handler = ExperimentJobApi(transport=self.transport, config=self.api_config)

    @httpretty.activate
    def test_get_job(self):
        obj = ExperimentJobConfig(uuid=uuid.uuid4().hex,
                                  experiment=1,
                                  created_at=datetime.datetime.now(),
                                  updated_at=datetime.datetime.now(),
                                  definition={}).to_dict()
        httpretty.register_uri(
            httpretty.GET,
            BaseApiHandler.build_url(
                self.api_config.base_url,
                '/',
                'username',
                'project_name',
                'experiments',
                1,
                'jobs',
                'uuid'),
            body=json.dumps(obj),
            content_type='application/json',
            status=200)

        # Schema response
        result = self.api_handler.get_job('username', 'project_name', 1, 'uuid')
        assert result.to_dict() == obj

        # Raw response
        self.set_raw_response()
        result = self.api_handler.get_job('username', 'project_name', 1, 'uuid')
        assert result == obj

    @httpretty.activate
    def test_get_experiment_job_status(self):
        obj = ExperimentJobStatusConfig(id=1,
                                        uuid=uuid.uuid4().hex,
                                        job=1,
                                        created_at=datetime.datetime.now(),
                                        status='Running').to_dict()
        httpretty.register_uri(
            httpretty.GET,
            BaseApiHandler.build_url(
                self.api_config.base_url,
                '/',
                'username',
                'project_name',
                'experiments',
                1,
                'jobs',
                1,
                'statuses'),
            body=json.dumps({'results': [obj], 'count': 1, 'next': None}),
            content_type='application/json',
            status=200)

        # Schema response
        response = self.api_handler.get_statuses('username', 'project_name', 1, 1)
        assert len(response['results']) == 1
        assert isinstance(response['results'][0], ExperimentJobStatusConfig)

        # Raw response
        self.set_raw_response()
        response = self.api_handler.get_statuses('username', 'project_name', 1, 1)
        assert len(response['results']) == 1
        assert isinstance(response['results'][0], Mapping)

    @httpretty.activate
    def test_experiment_job_logs(self):
        httpretty.register_uri(
            httpretty.GET,
            BaseApiHandler.build_url(
                self.api_config.base_url,
                '/',
                'username',
                'project_name',
                'experiments',
                1,
                'jobs',
                1,
                'logs'
            ),
            body='some text',
            content_type='text/plain',
            status=200)

        response = self.api_handler.logs('username', 'project_name', 1, 1, stream=False)
        assert response.content.decode() == 'some text'



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

import requests

import mock

from clint.textui.progress import Bar
from tests.test_transports.utils import BaseTestCaseTransport

from polyaxon_client.transport import Transport


class TestHttpTransport(BaseTestCaseTransport):
    # pylint:disable=protected-access
    def setUp(self):
        super(TestHttpTransport, self).setUp()
        self.transport = Transport()

    def test_create_progress_callback(self):
        encoder = mock.MagicMock()
        encoder.configure_mock(len=10)
        _, progress_bar = self.transport.create_progress_callback(encoder)
        assert isinstance(progress_bar, Bar)

    def test_format_sizeof(self):
        assert self.transport.format_sizeof(10) == '10.0B'
        assert self.transport.format_sizeof(10000) == '9.8KiB'
        assert self.transport.format_sizeof(100000) == '97.7KiB'
        assert self.transport.format_sizeof(10000000) == '9.5MiB'
        assert self.transport.format_sizeof(10000000000) == '9.3GiB'

    def test_session(self):
        assert hasattr(self.transport, '_session') is False
        assert isinstance(self.transport.session, requests.Session)
        assert isinstance(self.transport._session, requests.Session)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function


class PolyaxonClientException(Exception):
    def __init__(self, message=None):
        Exception.__init__(self, message)
        self.message = message

    def __str__(self):
        return self.message


class PolyaxonShouldExitError(PolyaxonClientException):
    pass


class PolyaxonHTTPError(PolyaxonClientException):
    def __init__(self, endpoint, response, message=None, status_code=None):
        super(PolyaxonHTTPError, self).__init__()
        self.endpoint = endpoint
        self.response = response
        self.message = getattr(self, 'message', message)
        self.status_code = getattr(self, 'status_code', status_code)

    def __str__(self):
        return '{status_code} on {endpoint}.'.format(status_code=self.status_code,
                                                     endpoint=self.endpoint)


class BadRequestError(PolyaxonHTTPError):
    status_code = 400
    message = "One or more request parameters is incorrect"


class AuthenticationError(PolyaxonHTTPError):
    staticmethod = 401
    message = "Authentication failed. Retry by invoking Polyaxon login."


class AuthorizationError(PolyaxonHTTPError):
    status_code = 403
    message = "You are not authorized to access this resource on Polyaxon."


class NotFoundError(PolyaxonHTTPError):
    status_code = 404
    message = "The resource you are looking for was not found. Check if the name or id is correct."


class OverLimitError(PolyaxonHTTPError):
    status_code = 429
    message = "You are over the allowed limits for this operation. Consider upgrading your account."


class ServerError(PolyaxonHTTPError):
    status_code = 500
    message = "Internal polyaxon server error, please try again later."


class BadGatewayError(PolyaxonHTTPError):
    status_code = 502
    message = "Invalid response from Polyaxon server."


class ServiceUnavailableError(PolyaxonHTTPError):
    status_code = 503
    message = "A problem was encountered, please try again later."


class GatewayTimeoutError(PolyaxonHTTPError):
    status_code = 504
    message = "Polyaxon server took too long to respond."


class SSLHandshakeError(PolyaxonHTTPError):
    status_code = 525


ERRORS_MAPPING = {
    'base': PolyaxonClientException,
    'http': PolyaxonHTTPError,
    400: BadRequestError,
    401: AuthenticationError,
    403: AuthorizationError,
    404: NotFoundError,
    429: OverLimitError,
    500: ServerError,
    502: BadGatewayError,
    503: ServiceUnavailableError,
    504: GatewayTimeoutError,
    525: SSLHandshakeError,
}



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from polyaxon_client.api.base import BaseApiHandler
from polyaxon_client.schemas import (
    ChartVersionConfig,
    CliVersionConfig,
    LibVersionConfig,
    LogHandlerConfig,
    PlatformVersionConfig
)


class VersionApi(BaseApiHandler):
    """
    Api handler to get API version from the server.
    """
    ENDPOINT = "/versions/"

    def get_cli_version(self):
        request_url = self.build_url(self._get_http_url(), 'cli')
        response = self.transport.get(request_url)
        return self.prepare_results(response_json=response.json(), config=CliVersionConfig)

    def get_platform_version(self):
        request_url = self.build_url(self._get_http_url(), 'platform')
        response = self.transport.get(request_url)
        return self.prepare_results(response_json=response.json(), config=PlatformVersionConfig)

    def get_lib_version(self):
        request_url = self.build_url(self._get_http_url(), 'lib')
        response = self.transport.get(request_url)
        return self.prepare_results(response_json=response.json(), config=LibVersionConfig)

    def get_chart_version(self):
        request_url = self.build_url(self._get_http_url(), 'chart')
        response = self.transport.get(request_url)
        return self.prepare_results(response_json=response.json(), config=ChartVersionConfig)

    def get_log_handler(self):
        request_url = self.build_url(self._get_http_url('/log_handler'))
        response = self.transport.get(request_url)
        return self.prepare_results(response_json=response.json(), config=LogHandlerConfig)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

import uuid

from tests.utils import TestEnvVarsCase

from polyaxon_client import settings
from polyaxon_client.exceptions import PolyaxonClientException
from polyaxon_client.tracking.job import Job


class TestJobTracking(TestEnvVarsCase):
    def setUp(self):
        super(TestJobTracking, self).setUp()
        settings.IN_CLUSTER = True

    def test_job_info_checks_in_cluster(self):
        settings.IN_CLUSTER = False
        with self.assertRaises(PolyaxonClientException):
            Job.get_job_info()

    def test_empty_job_info(self):
        self.check_empty_value('POLYAXON_JOB_INFO', Job.get_job_info)

    def test_non_dict_job_info(self):
        self.check_non_dict_value('POLYAXON_JOB_INFO', Job.get_job_info)

    def test_dict_job_info(self):
        job_info = {
            "project_name": "project_bar",
            "job_name": "project_bar.jobs.1",
            "project_uuid": uuid.uuid4().hex,
            "job_uuid": uuid.uuid4().hex,
        }
        self.check_valid_dict_value('POLYAXON_JOB_INFO',
                                    Job.get_job_info,
                                    job_info)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from polyaxon_client.exceptions import PolyaxonClientException


def get_project_info(current_user, project):
    if not project:
        raise PolyaxonClientException('You need to provide a project to do tracking.')
    parts = project.replace('.', '/').split('/')
    if len(parts) == 2:
        user, project_name = parts
    else:
        user = current_user
        project_name = project

    return user, project_name



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from tests.test_api.utils import TestBaseApi

from polyaxon_client.api.base import BaseApiHandler
from polyaxon_client.exceptions import ERRORS_MAPPING


class DummyConfig(object):
    @staticmethod
    def from_dict(value, unknown=None):  # noqa
        return 'from_dict', value


class TestBaseApiHandler(TestBaseApi):
    # pylint:disable=protected-access
    def setUp(self):
        super(TestBaseApiHandler, self).setUp()
        self.api_handler = BaseApiHandler(transport=self.transport, config=self.api_config)

    def test_get_page(self):
        assert self.api_handler.get_page() == {}
        assert self.api_handler.get_page(page=1) == {}
        assert self.api_handler.get_page(page=2) == {'offset': self.api_config.PAGE_SIZE}
        assert self.api_handler.get_page(page=3) == {'offset': self.api_config.PAGE_SIZE * 2}

    def test_build_url(self):
        assert self.api_handler.build_url('a') == 'a/'
        assert self.api_handler.build_url('a', 'b') == 'a/b/'

    def test_get_url(self):
        with self.assertRaises(ERRORS_MAPPING['base']):
            self.api_handler._get_url('a')

        assert self.api_handler._get_url('base', 'endpoint') == 'base/endpoint/'
        assert self.api_handler._get_http_url('endpoint') == '{}/endpoint/'.format(
            self.api_config.base_url)
        assert self.api_handler._get_ws_url('endpoint') == '{}/endpoint/'.format(
            self.api_config.base_ws_url)

    def test_prepare_results(self):
        # Schema response
        results = self.api_handler.prepare_results(response_json={'foo': 'bar'},
                                                   config=DummyConfig)
        assert results == ('from_dict', {'foo': 'bar'})

        # Raw response
        self.api_config.schema_response = False
        results = self.api_handler.prepare_results(response_json={'foo': 'bar'},
                                                   config=DummyConfig)
        assert results == {'foo': 'bar'}

    def test_prepare_list_results(self):
        # Schema response
        results = self.api_handler.prepare_list_results(response_json={'results': ['bar']},
                                                        current_page=1,
                                                        config=DummyConfig)
        assert results == {'count': 0,
                           'next': None,
                           'previous': None,
                           'results': [('from_dict', 'bar')]}

        # Raw response
        self.api_config.schema_response = False
        results = self.api_handler.prepare_list_results(response_json={'results': ['bar']},
                                                        current_page=1,
                                                        config=DummyConfig)
        assert results == {'count': 0, 'next': None, 'previous': None, 'results': ['bar']}



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from polyaxon_client.exceptions import AuthenticationError
from polyaxon_client.logger import logger
from polyaxon_client.transport.http_transport import HttpTransportMixin
from polyaxon_client.transport.periodic_transport import (
    PeriodicHttpTransportMixin,
    PeriodicWSTransportMixin
)
from polyaxon_client.transport.socket_transport import SocketTransportMixin
from polyaxon_client.transport.threaded_transport import ThreadedTransportMixin


class Transport(HttpTransportMixin,
                PeriodicHttpTransportMixin,
                PeriodicWSTransportMixin,
                ThreadedTransportMixin,
                SocketTransportMixin):
    """Transport for handling http/ws operations."""

    def __init__(self, config=None):
        self.config = config

    def _get_headers(self, headers=None):
        request_headers = headers or {}
        # Auth headers if access_token is present
        if self.config:
            if 'Authorization' not in request_headers and self.config.token:
                request_headers.update({'Authorization': '{} {}'.format(
                    self.config.authentication_type, self.config.token)})
            if self.config.internal_header:
                request_headers.update(self.config.internal_header)
        return request_headers

    def handle_exception(self, e, log_message=None):
        logger.info("%s: %s", log_message, e.message)

        if self.config.reraise:
            raise e

        if isinstance(e, AuthenticationError):
            # exit now since there is nothing we can do without login
            raise e



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

import logging
import sys

logger = logging.getLogger('polyaxon.client')


def configure_logger(verbose):
    log_level = logging.DEBUG if verbose else logging.INFO
    logging.basicConfig(format='%(levelname)s: %(asctime)s %(message)s',
                        level=log_level,
                        stream=sys.stdout)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

import requests
import time

from flaky import flaky
from tests.test_transports.utils import BaseTestCaseTransport

from polyaxon_client.api_config import ApiConfig
from polyaxon_client.transport.threaded_transport import ThreadedTransportMixin
from polyaxon_client.workers.queue_worker import QueueWorker


class DummyTransport(ThreadedTransportMixin):
    # pylint:disable=protected-access
    def __init__(self, delay=0):
        self.queue = []
        self.delay = delay
        self.config = ApiConfig(in_cluster=True, timeout=0.01)
        self._threaded_exceptions = 0
        self._threaded_done = 0

    def post(self, url, **kwargs):
        time.sleep(self.delay)
        self.queue.append(('post', url))

    def patch(self, url, **kwargs):
        time.sleep(self.delay)
        self.queue.append(('patch', url))

    def delete(self, url, **kwargs):
        time.sleep(self.delay)
        self.queue.append(('delete', url))

    def put(self, url, **kwargs):
        time.sleep(self.delay)
        self.queue.append(('put', url))

    def upload(self, url, **kwargs):
        time.sleep(self.delay)
        self.queue.append(('upload', url))


class ExceptionTransport(ThreadedTransportMixin):
    # pylint:disable=protected-access
    def __init__(self, delay=0):
        self.delay = delay
        self.config = ApiConfig(in_cluster=True, timeout=0.01)
        self._threaded_exceptions = 0
        self._threaded_done = 0

    def post(self, **kwargs):
        time.sleep(self.delay)
        raise requests.exceptions.HTTPError('error')

    def patch(self, **kwargs):
        time.sleep(self.delay)
        raise requests.exceptions.HTTPError('error')

    def delete(self, **kwargs):
        time.sleep(self.delay)
        raise requests.exceptions.HTTPError('error')

    def put(self, **kwargs):
        time.sleep(self.delay)
        raise requests.exceptions.HTTPError('error')

    def upload(self, **kwargs):
        time.sleep(self.delay)
        raise requests.exceptions.HTTPError('error')


class TestThreadedTransport(BaseTestCaseTransport):
    # pylint:disable=protected-access
    def setUp(self):
        super(TestThreadedTransport, self).setUp()
        self.transport = DummyTransport()
        self.exception_transport = ExceptionTransport()

    def test_retry_session(self):
        assert hasattr(self.transport, '_retry_session') is False
        assert isinstance(self.transport.retry_session, requests.Session)
        assert isinstance(self.transport._retry_session, requests.Session)

    def test_worker(self):
        assert hasattr(self.transport, '_worker') is False
        assert isinstance(self.transport.worker, QueueWorker)
        assert isinstance(self.transport._worker, QueueWorker)

    def test_async_requests(self):
        assert self.transport.queue == []

        self.transport.async_post(url='url_post')
        time.sleep(0.03)
        assert self.transport.queue == [('post', 'url_post')]
        assert self.transport.threaded_done == 1
        assert self.transport.threaded_exceptions == 0

        self.transport.async_patch(url='url_patch')
        time.sleep(0.03)
        assert self.transport.queue == [('post', 'url_post'), ('patch', 'url_patch')]
        assert self.transport.threaded_done == 2
        assert self.transport.threaded_exceptions == 0

        self.transport.async_put(url='url_put')
        time.sleep(0.03)
        assert self.transport.queue == [('post', 'url_post'),
                                        ('patch', 'url_patch'),
                                        ('put', 'url_put')]
        assert self.transport.threaded_done == 3
        assert self.transport.threaded_exceptions == 0

        self.transport.async_delete(url='url_delete')
        time.sleep(0.03)
        assert self.transport.queue == [('post', 'url_post'),
                                        ('patch', 'url_patch'),
                                        ('put', 'url_put'),
                                        ('delete', 'url_delete')]
        assert self.transport.threaded_done == 4
        assert self.transport.threaded_exceptions == 0

        self.transport.async_upload(url='url_upload', files=['file'], files_size=200)
        time.sleep(0.03)
        assert self.transport.queue == [('post', 'url_post'),
                                        ('patch', 'url_patch'),
                                        ('put', 'url_put'),
                                        ('delete', 'url_delete'),
                                        ('upload', 'url_upload')]
        assert self.transport.threaded_done == 5
        assert self.transport.threaded_exceptions == 0
        assert self.transport.worker.is_alive() is True

    @flaky(max_runs=3)
    def test_async_exceptions(self):
        self.exception_transport.async_post(url='url_post')
        time.sleep(0.03)
        assert self.exception_transport.threaded_done == 1
        assert self.exception_transport.threaded_exceptions == 1

        self.exception_transport.async_patch(url='url_patch')
        time.sleep(0.03)
        assert self.exception_transport.threaded_done == 2
        assert self.exception_transport.threaded_exceptions == 2

        self.exception_transport.async_put(url='url_put')
        time.sleep(0.03)
        assert self.exception_transport.threaded_done == 3
        assert self.exception_transport.threaded_exceptions == 3

        self.exception_transport.async_delete(url='url_delete')
        time.sleep(0.03)
        assert self.exception_transport.threaded_done == 4
        assert self.exception_transport.threaded_exceptions == 4

        self.exception_transport.async_delete(url='url_upload')
        time.sleep(0.03)
        assert self.exception_transport.threaded_done == 5
        assert self.exception_transport.threaded_exceptions == 5

    def test_worker_atexit_handle_queue_before_stopping(self):
        # Transport
        self.transport.config.timeout = 0.5
        self.transport.delay = 0.5
        assert self.transport.queue == []
        self.transport.async_post(url='url_post')
        assert self.transport.queue == []
        time.sleep(0.1)
        assert self.transport.worker.is_alive() is True
        self.transport.worker.atexit()
        assert self.transport.queue == [('post', 'url_post')]
        assert self.transport.threaded_done == 1
        assert self.transport.threaded_exceptions == 0
        assert self.transport._worker.is_alive() is False

        # Exception transport
        self.exception_transport.config.timeout = 0.5
        self.exception_transport.delay = 0.5
        self.exception_transport.async_post(url='url_post')
        assert self.exception_transport.threaded_done == 0
        assert self.exception_transport.threaded_exceptions == 0
        time.sleep(0.1)
        assert self.exception_transport.worker.is_alive() is True
        self.exception_transport.worker.atexit()
        assert self.exception_transport.threaded_done == 1
        assert self.exception_transport.threaded_exceptions == 1
        assert self.exception_transport._worker.is_alive() is False



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

import numpy as np

from hestia.tz_utils import utc

from polyaxon_client.api.base import BaseApiHandler
from polyaxon_client.exceptions import PolyaxonClientException
from polyaxon_client.schemas import (
    ExperimentConfig,
    ExperimentJobConfig,
    ExperimentMetricConfig,
    ExperimentStatusConfig,
    CodeReferenceConfig
)


class ExperimentApi(BaseApiHandler):
    """
    Api handler to get experiments from the server.
    """
    ENDPOINT = "/"

    def list_experiments(self, page=1):
        """This gets all experiments visible to the user from the server."""
        try:
            response = self.transport.get(self._get_http_url('/experiments'),
                                          params=self.get_page(page=page))
            return self.prepare_list_results(response.json(), page, ExperimentConfig)
        except PolyaxonClientException as e:
            self.transport.handle_exception(e=e, log_message='Error while retrieving experiments.')
            return []

    def get_experiment(self, username, project_name, experiment_id):
        request_url = self.build_url(self._get_http_url(),
                                     username,
                                     project_name,
                                     'experiments',
                                     experiment_id)
        try:
            response = self.transport.get(request_url)
            return self.prepare_results(response_json=response.json(), config=ExperimentConfig)
        except PolyaxonClientException as e:
            self.transport.handle_exception(e=e, log_message='Error while retrieving experiment.')
            return None

    def update_experiment(self,
                          username,
                          project_name,
                          experiment_id,
                          patch_dict,
                          background=False):
        request_url = self.build_url(self._get_http_url(),
                                     username,
                                     project_name,
                                     'experiments',
                                     experiment_id)

        if background:
            self.transport.async_patch(request_url, json_data=patch_dict)
            return None

        try:
            response = self.transport.patch(request_url, json_data=patch_dict)
            return self.prepare_results(response_json=response.json(), config=ExperimentConfig)
        except PolyaxonClientException as e:
            self.transport.handle_exception(e=e, log_message='Error while updating experiment.')
            return None

    def delete_experiment(self, username, project_name, experiment_id, background=False):
        request_url = self.build_url(self._get_http_url(),
                                     username,
                                     project_name,
                                     'experiments',
                                     experiment_id)

        if background:
            self.transport.async_delete(request_url)
            return None

        try:
            return self.transport.delete(request_url)
        except PolyaxonClientException as e:
            self.transport.handle_exception(e=e, log_message='Error while deleting experiment.')
            return None

    def get_statuses(self, username, project_name, experiment_id, page=1):
        request_url = self.build_url(self._get_http_url(),
                                     username,
                                     project_name,
                                     'experiments',
                                     experiment_id,
                                     'statuses')
        try:
            response = self.transport.get(request_url, params=self.get_page(page=page))
            return self.prepare_list_results(response.json(), page, ExperimentStatusConfig)
        except PolyaxonClientException as e:
            self.transport.handle_exception(
                e=e, log_message='Error while retrieving experiment statuses.')
            return None

    def create_status(self,
                      username,
                      project_name,
                      experiment_id,
                      status,
                      message=None,
                      traceback=None,
                      background=False):
        request_url = self.build_url(self._get_http_url(),
                                     username,
                                     project_name,
                                     'experiments',
                                     experiment_id,
                                     'statuses')

        json_data = {'status': status}
        if message:
            json_data['message'] = message
        if traceback:
            json_data['traceback'] = traceback
        if background:
            self.transport.async_post(request_url, json_data=json_data)
            return None

        try:
            response = self.transport.post(request_url, json_data=json_data)
            return self.prepare_results(response_json=response.json(),
                                        config=ExperimentStatusConfig)
        except PolyaxonClientException as e:
            self.transport.handle_exception(
                e=e, log_message='Error while creating experiment status.')
            return None

    def get_code_reference(self,
                           username,
                           project_name,
                           experiment_id):
        request_url = self.build_url(self._get_http_url(),
                                     username,
                                     project_name,
                                     'experiments',
                                     experiment_id,
                                     'coderef')
        try:
            response = self.transport.get(request_url)
            return self.prepare_results(response.json(), CodeReferenceConfig)
        except PolyaxonClientException as e:
            self.transport.handle_exception(
                e=e, log_message='Error while retrieving experiment code reference.')
            return None

    def create_code_reference(self,
                              username,
                              project_name,
                              experiment_id,
                              coderef,
                              background=False):
        request_url = self.build_url(self._get_http_url(),
                                     username,
                                     project_name,
                                     'experiments',
                                     experiment_id,
                                     'coderef')
        if background:
            self.transport.async_post(request_url, json_data=coderef)
            return None

        try:
            response = self.transport.post(request_url, json_data=coderef)
            return self.prepare_results(response_json=response.json(),
                                        config=CodeReferenceConfig)
        except PolyaxonClientException as e:
            self.transport.handle_exception(
                e=e, log_message='Error while creating experiment coderef.')
            return None

    def get_metrics(self, username, project_name, experiment_id, page=1):
        request_url = self.build_url(self._get_http_url(),
                                     username,
                                     project_name,
                                     'experiments',
                                     experiment_id,
                                     'metrics')
        try:
            response = self.transport.get(request_url, params=self.get_page(page=page))
            return self.prepare_list_results(response.json(), page, ExperimentMetricConfig)
        except PolyaxonClientException as e:
            self.transport.handle_exception(
                e=e, log_message='Error while retrieving experiment metric.')
            return None

    def create_metric(self,
                      username,
                      project_name,
                      experiment_id,
                      values,
                      created_at=None,
                      background=False,
                      periodic=False):
        request_url = self.build_url(self._get_http_url(),
                                     username,
                                     project_name,
                                     'experiments',
                                     experiment_id,
                                     'metrics')

        # Validate metric values
        def parse_numbers(value):
            if isinstance(value, (int, float, complex, type(None))):
                return value
            if isinstance(value, np.integer):
                return int(value)
            if isinstance(value, np.floating):
                return float(value)
            raise PolyaxonClientException(
                'Client could not parse the value `{}`, it expects a number.'.format(value))

        json_data = {'values': {key: parse_numbers(values[key]) for key in values}}

        if created_at:
            json_data['created_at'] = str(utc.localize(created_at))

        if background:
            self.transport.async_post(request_url, json_data=json_data)
            return None

        if periodic:
            self.transport.periodic_post(request_url, json_data=json_data)
            return None

        try:
            response = self.transport.post(request_url, json_data=json_data)
            return self.prepare_results(response_json=response.json(),
                                        config=ExperimentMetricConfig)
        except PolyaxonClientException as e:
            self.transport.handle_exception(
                e=e, log_message='Error while retrieving experiment status.')
            return None

    def send_logs(self,
                  username,
                  project_name,
                  experiment_id,
                  log_lines,
                  background=False,
                  periodic=False):
        request_url = self.build_url(self._get_http_url(),
                                     username,
                                     project_name,
                                     'experiments',
                                     experiment_id,
                                     'logs')

        if isinstance(log_lines, list):
            log_lines = '\n'.join(log_lines)

        if background:
            self.transport.async_post(request_url, json_data=log_lines)
            return None

        if periodic:
            self.transport.periodic_post(request_url, json_data=log_lines)
            return None

        try:
            response = self.transport.post(request_url, json_data=log_lines)
            return response
        except PolyaxonClientException as e:
            self.transport.handle_exception(
                e=e, log_message='Error while retrieving experiment status.')
            return None

    def list_jobs(self, username, project_name, experiment_id, page=1):
        """Fetch list of jobs related to this experiment."""
        request_url = self.build_url(self._get_http_url(),
                                     username,
                                     project_name,
                                     'experiments',
                                     experiment_id,
                                     'jobs')

        try:
            response = self.transport.get(request_url, params=self.get_page(page=page))
            return self.prepare_list_results(response.json(), page, ExperimentJobConfig)
        except PolyaxonClientException as e:
            self.transport.handle_exception(e=e, log_message='Error while retrieving jobs.')
            return []

    def restart(self,
                username,
                project_name,
                experiment_id,
                config=None,
                update_code=None,
                background=False):
        """Restart an experiment."""
        request_url = self.build_url(self._get_http_url(),
                                     username,
                                     project_name,
                                     'experiments',
                                     experiment_id,
                                     'restart')

        data = {}
        if config:
            data['config'] = config
        if update_code:
            data['update_code'] = update_code

        if background:
            self.transport.async_post(request_url, json_data=data)
            return None

        try:
            response = self.transport.post(request_url, json_data=data)
            return self.prepare_results(response_json=response.json(), config=ExperimentConfig)
        except PolyaxonClientException as e:
            self.transport.handle_exception(
                e=e, log_message='Error while restarting the experiment.')
            return None

    def resume(self,
               username,
               project_name,
               experiment_id,
               config=None,
               update_code=None,
               background=False):
        """Restart an experiment."""
        request_url = self.build_url(self._get_http_url(),
                                     username,
                                     project_name,
                                     'experiments',
                                     experiment_id,
                                     'resume')

        data = {}
        if config:
            data['config'] = config
        if update_code:
            data['update_code'] = update_code

        if background:
            self.transport.async_post(request_url, json_data=data)
            return None

        try:
            response = self.transport.post(request_url, json_data=data)
            return self.prepare_results(response_json=response.json(), config=ExperimentConfig)
        except PolyaxonClientException as e:
            self.transport.handle_exception(e=e, log_message='Error while resuming the experiment.')
            return None

    def copy(self,
             username,
             project_name,
             experiment_id,
             config=None,
             update_code=None,
             background=False):
        """Restart an experiment."""
        request_url = self.build_url(self._get_http_url(),
                                     username,
                                     project_name,
                                     'experiments',
                                     experiment_id,
                                     'copy')

        data = {}
        if config:
            data['config'] = config
        if update_code:
            data['update_code'] = update_code

        if background:
            self.transport.async_post(request_url, json_data=data)
            return None

        try:
            response = self.transport.post(request_url, json_data=data)
            return self.prepare_results(response_json=response.json(), config=ExperimentConfig)
        except PolyaxonClientException as e:
            self.transport.handle_exception(e=e, log_message='Error while copying the experiment.')
            return None

    def stop(self, username, project_name, experiment_id, background=False):
        request_url = self.build_url(self._get_http_url(),
                                     username,
                                     project_name,
                                     'experiments',
                                     experiment_id,
                                     'stop')

        if background:
            self.transport.async_post(request_url)
            return None

        try:
            return self.transport.post(request_url)
        except PolyaxonClientException as e:
            self.transport.handle_exception(e=e, log_message='Error while stopping experiment.')
            return None

    def resources(self, username, project_name, experiment_id, message_handler=None):
        """Streams experiments resources using websockets.

        message_handler: handles the messages received from server.
            e.g. def f(x): print(x)
        """
        request_url = self.build_url(self._get_ws_url(),
                                     username,
                                     project_name,
                                     'experiments',
                                     experiment_id,
                                     'resources')
        self.transport.stream(request_url, message_handler=message_handler)

    # pylint:disable=inconsistent-return-statements
    def logs(self, username, project_name, experiment_id, stream=True, message_handler=None):
        """Streams experiments logs using websockets.

        message_handler: handles the messages received from server.
            e.g. def f(x): print(x)
        """
        if not stream:
            request_url = self.build_url(self._get_http_url(),
                                         username,
                                         project_name,
                                         'experiments',
                                         experiment_id,
                                         'logs')

            try:
                return self.transport.get(request_url)
            except PolyaxonClientException as e:
                self.transport.handle_exception(
                    e=e, log_message='Error while retrieving experiment logs.')
                return []

        request_url = self.build_url(self._get_ws_url(),
                                     username,
                                     project_name,
                                     'experiments',
                                     experiment_id,
                                     'logs')
        self.transport.stream(request_url, message_handler=message_handler)

    def start_tensorboard(self,
                          username,
                          project_name,
                          experiment_id,
                          job_config=None,
                          background=False):
        request_url = self.build_url(self._get_http_url(),
                                     username,
                                     project_name,
                                     'experiments',
                                     experiment_id,
                                     'tensorboard',
                                     'start')

        job_config = {'config': job_config} if job_config else {}

        if background:
            self.transport.async_post(request_url, json_data=job_config)
            return None

        try:
            return self.transport.post(request_url, json_data=job_config)
        except PolyaxonClientException as e:
            self.transport.handle_exception(e=e, log_message='Error while starting tensorboard.')
            return None

    def stop_tensorboard(self, username, project_name, experiment_id, background=False):
        request_url = self.build_url(self._get_http_url(),
                                     username,
                                     project_name,
                                     'experiments',
                                     experiment_id,
                                     'tensorboard',
                                     'stop')

        if background:
            self.transport.async_post(request_url)
            return None

        try:
            return self.transport.post(request_url)
        except PolyaxonClientException as e:
            self.transport.handle_exception(e=e, log_message='Error while stopping tensorboard.')
            return None

    def bookmark(self, username, project_name, experiment_id, background=False):
        request_url = self.build_url(self._get_http_url(),
                                     username,
                                     project_name,
                                     'experiments',
                                     experiment_id,
                                     'bookmark')

        if background:
            self.transport.async_post(request_url)
            return None

        try:
            return self.transport.post(request_url)
        except PolyaxonClientException as e:
            self.transport.handle_exception(e=e, log_message='Error while bookmarking experiment.')
            return None

    def unbookmark(self, username, project_name, experiment_id, background=False):
        request_url = self.build_url(self._get_http_url(),
                                     username,
                                     project_name,
                                     'experiments',
                                     experiment_id,
                                     'unbookmark')

        if background:
            self.transport.async_delete(request_url)
            return None

        try:
            return self.transport.delete(request_url)
        except PolyaxonClientException as e:
            self.transport.handle_exception(
                e=e, log_message='Error while unbookmarking experiment.')
            return None

    def download_outputs(self, username, project_name, experiment_id):
        """Downloads outputs for this experiment to the current dir."""
        request_url = self.build_url(self._get_http_url(),
                                     username,
                                     project_name,
                                     'experiments',
                                     experiment_id,
                                     'outputs',
                                     'download')

        try:
            response = self.transport.download(
                request_url,
                '{}.{}.{}.tar.gz'.format(username, project_name, experiment_id))
            return response
        except PolyaxonClientException as e:
            self.transport.handle_exception(
                e=e, log_message='Error while downloading experiment outputs.')
            return None

    def get_heartbeat_url(self, username, project_name, experiment_id):
        return self.build_url(self._get_http_url(),
                              username,
                              project_name,
                              'experiments',
                              experiment_id,
                              self.HEARTBEAT)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from tests.utils import TestEnvVarsCase

from polyaxon_client import settings
from polyaxon_client.exceptions import PolyaxonClientException
from polyaxon_client.tracking import paths


class TestTracker(TestEnvVarsCase):
    def test_empty_base_outputs_path(self):
        self.check_empty_value('POLYAXON_BASE_OUTPUTS_PATH', paths.get_base_outputs_path)

    def test_valid_base_outputs_path(self):
        self.check_valid_value('POLYAXON_BASE_OUTPUTS_PATH', paths.get_base_outputs_path, 'path')

    def test_get_outputs_raises_out_cluster(self):
        settings.IN_CLUSTER = False
        with self.assertRaises(PolyaxonClientException):
            paths.get_outputs_path()

    def test_empty_outputs_path(self):
        settings.IN_CLUSTER = True
        self.check_empty_value('POLYAXON_RUN_OUTPUTS_PATH', paths.get_outputs_path)

    def test_valid_outputs_path(self):
        settings.IN_CLUSTER = True
        self.check_valid_value('POLYAXON_RUN_OUTPUTS_PATH', paths.get_outputs_path, 'path')

    def test_get_data_paths_raises_out_cluster(self):
        settings.IN_CLUSTER = False
        with self.assertRaises(PolyaxonClientException):
            paths.get_data_paths()

    def test_empty_data_path(self):
        settings.IN_CLUSTER = True
        self.check_empty_value('POLYAXON_RUN_DATA_PATHS', paths.get_data_paths)

    def test_valid_data_path(self):
        settings.IN_CLUSTER = True
        self.check_valid_dict_value('POLYAXON_RUN_DATA_PATHS', paths.get_data_paths,
                                    {'data': 'path'})

    def test_get_outputs_refs_paths_raises_out_cluster(self):
        settings.IN_CLUSTER = False
        with self.assertRaises(PolyaxonClientException):
            paths.get_outputs_refs_paths()

    def test_empty_outputs_refs_paths(self):
        settings.IN_CLUSTER = True
        self.check_empty_value('POLYAXON_REFS_OUTPUTS_PATHS', paths.get_outputs_refs_paths)

    def test_valid_outputs_refs_paths(self):
        settings.IN_CLUSTER = True
        self.check_valid_dict_value('POLYAXON_REFS_OUTPUTS_PATHS',
                                    paths.get_outputs_refs_paths,
                                    {
                                        'jobs': ['path1', 'path12'],
                                        'experiments': ['path1', 'path12']
                                    })

    def test_get_log_level_raises_out_cluster(self):
        settings.IN_CLUSTER = False
        with self.assertRaises(PolyaxonClientException):
            paths.get_log_level()

    def test_empty_log_level(self):
        settings.IN_CLUSTER = True
        self.check_empty_value('POLYAXON_LOG_LEVEL', paths.get_log_level)

    def test_valid_log_level(self):
        settings.IN_CLUSTER = True
        self.check_valid_value('POLYAXON_LOG_LEVEL', paths.get_log_level, 'info')



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

import os
import shlex

from subprocess import PIPE

from psutil import Popen


def run_command(cmd, data, location, chw):
    cwd = os.getcwd()
    if location is not None and chw is True:
        cwd = location
    elif location is not None and chw is False:
        cmd = '{0} {1}'.format(cmd, location)
    r = Popen(shlex.split(cmd), stdout=PIPE, stdin=PIPE, stderr=PIPE, cwd=cwd)
    if data is None:
        output = r.communicate()[0].decode('utf-8')
    else:
        output = r.communicate(input=data)[0]
    return output



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from polyaxon_client.api.base import BaseApiHandler
from polyaxon_client.exceptions import PolyaxonClientException
from polyaxon_client.schemas import ExperimentConfig, ExperimentGroupConfig, GroupStatusConfig


class ExperimentGroupApi(BaseApiHandler):
    """
    Api handler to get experiments for a group from the server.
    """
    ENDPOINT = "/"

    def get_experiment_group(self, username, project_name, group_id):
        request_url = self.build_url(self._get_http_url(),
                                     username,
                                     project_name,
                                     'groups',
                                     group_id)
        try:
            response = self.transport.get(request_url)
            return self.prepare_results(response_json=response.json(), config=ExperimentGroupConfig)
        except PolyaxonClientException as e:
            self.transport.handle_exception(e=e, log_message='Error while retrieving project')
            return None

    def list_experiments(self,
                         username,
                         project_name,
                         group_id,
                         metrics=None,
                         declarations=None,
                         query=None,
                         sort=None,
                         page=1):
        """Fetch list of experiments related to this experiment group."""
        request_url = self.build_url(self._get_http_url(), username, project_name, 'experiments')

        try:
            params = self.get_page(page=page)
            params['group'] = group_id
            if metrics:
                params['metrics'] = metrics
            if declarations:
                params['declarations'] = declarations
            if query:
                params['query'] = query
            if sort:
                params['sort'] = sort
            response = self.transport.get(request_url, params=params)
            return self.prepare_list_results(response.json(), page, ExperimentConfig)
        except PolyaxonClientException as e:
            self.transport.handle_exception(e=e, log_message='Error while retrieving experiments')
            return []

    def update_experiment_group(self,
                                username,
                                project_name,
                                group_id,
                                patch_dict,
                                background=False):
        request_url = self.build_url(self._get_http_url(),
                                     username,
                                     project_name,
                                     'groups',
                                     group_id)

        if background:
            self.transport.async_patch(request_url, json_data=patch_dict)
            return None

        try:
            response = self.transport.patch(request_url, json_data=patch_dict)
            return self.prepare_results(response_json=response.json(), config=ExperimentGroupConfig)
        except PolyaxonClientException as e:
            self.transport.handle_exception(e=e, log_message='Error while updating project')
            return None

    def delete_experiment_group(self, username, project_name, group_id, background=False):
        request_url = self.build_url(self._get_http_url(),
                                     username,
                                     project_name,
                                     'groups',
                                     group_id)

        if background:
            self.transport.async_delete(request_url)
            return None

        try:
            response = self.transport.delete(request_url)
            return response
        except PolyaxonClientException as e:
            self.transport.handle_exception(
                e=e, log_message='Error while deleting experiment group')
            return None

    def get_statuses(self, username, project_name, group_id, page=1):
        request_url = self.build_url(self._get_http_url(),
                                     username,
                                     project_name,
                                     'groups',
                                     group_id,
                                     'statuses')
        try:
            response = self.transport.get(request_url, params=self.get_page(page=page))
            return self.prepare_list_results(response.json(), page, GroupStatusConfig)
        except PolyaxonClientException as e:
            self.transport.handle_exception(
                e=e, log_message='Error while retrieving group statuses')
            return None

    def create_status(self,
                      username,
                      project_name,
                      group_id,
                      status,
                      message=None,
                      background=False):
        request_url = self.build_url(self._get_http_url(),
                                     username,
                                     project_name,
                                     'groups',
                                     group_id,
                                     'statuses')

        json_data = {'status': status}
        if message:
            json_data['message'] = message
        if background:
            self.transport.async_post(request_url, json_data=json_data)
            return None

        try:
            response = self.transport.post(request_url, json_data=json_data)
            return self.prepare_results(response_json=response.json(),
                                        config=GroupStatusConfig)
        except PolyaxonClientException as e:
            self.transport.handle_exception(
                e=e, log_message='Error while creating group status.')
            return None

    def stop(self, username, project_name, group_id, pending=False, background=False):
        request_url = self.build_url(self._get_http_url(),
                                     username,
                                     project_name,
                                     'groups',
                                     group_id,
                                     'stop')
        json_data = None
        if pending is True:
            json_data = {'pending': pending}

        if background:
            self.transport.async_post(request_url, json_data=json_data)
            return None

        try:
            return self.transport.post(request_url, json_data=json_data)
        except PolyaxonClientException as e:
            self.transport.handle_exception(
                e=e, log_message='Error while stopping experiments in group')
            return None

    def start_tensorboard(self,
                          username,
                          project_name,
                          group_id,
                          job_config=None,
                          background=False):
        request_url = self.build_url(self._get_http_url(),
                                     username,
                                     project_name,
                                     'groups',
                                     group_id,
                                     'tensorboard',
                                     'start')

        job_config = {'config': job_config} if job_config else {}

        if background:
            self.transport.async_post(request_url, json_data=job_config)
            return None

        try:
            return self.transport.post(request_url, json_data=job_config)
        except PolyaxonClientException as e:
            self.transport.handle_exception(e=e, log_message='Error while starting tensorboard')
            return None

    def stop_tensorboard(self, username, project_name, group_id, background=False):
        request_url = self.build_url(self._get_http_url(),
                                     username,
                                     project_name,
                                     'groups',
                                     group_id,
                                     'tensorboard',
                                     'stop')

        if background:
            self.transport.async_post(request_url)
            return None

        try:
            return self.transport.post(request_url)
        except PolyaxonClientException as e:
            self.transport.handle_exception(e=e, log_message='Error while stopping tensorboard')
            return None

    def bookmark(self, username, project_name, group_id, background=False):
        request_url = self.build_url(self._get_http_url(),
                                     username,
                                     project_name,
                                     'groups',
                                     group_id,
                                     'bookmark')

        if background:
            self.transport.async_post(request_url)
            return None

        try:
            return self.transport.post(request_url)
        except PolyaxonClientException as e:
            self.transport.handle_exception(e=e, log_message='Error while bookmarking group')
            return None

    def unbookmark(self, username, project_name, group_id, background=False):
        request_url = self.build_url(self._get_http_url(),
                                     username,
                                     project_name,
                                     'groups',
                                     group_id,
                                     'unbookmark')

        if background:
            self.transport.async_delete(request_url)
            return None

        try:
            return self.transport.delete(request_url)
        except PolyaxonClientException as e:
            self.transport.handle_exception(e=e, log_message='Error while unbookmarking group')
            return None



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

import atexit
import os
import threading


class BaseWorker(object):
    NAME = None

    def __init__(self):
        assert self.NAME, 'Worker class `{}` must have a valid name.'.format(
            self.__class__.__name__)
        self._lock = threading.Lock()
        self._thread = None
        self._thread_for_pid = None

    def is_alive(self):
        if self._thread_for_pid != os.getpid():
            return False
        return bool(self._thread and self._thread.is_alive())

    def is_running(self):
        if self.is_alive():
            return
        self.start()

    def start(self):
        self._lock.acquire()
        try:
            if not self.is_alive():
                self._thread = threading.Thread(target=self._target, name=self.NAME)
                self._thread.setDaemon(True)
                self._thread.start()
                self._thread_for_pid = os.getpid()
        finally:
            self._lock.release()
            atexit.register(self.atexit)

    def atexit(self):
        raise NotImplementedError('Worker must implement `atexit` function.')

    def _target(self):
        raise NotImplementedError('Worker must implement `target` function.')



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

import json
import os
import uuid

from tests.utils import TestEnvVarsCase

from polyaxon_client import settings
from polyaxon_client.exceptions import PolyaxonClientException
from polyaxon_client.tracking.experiment import Experiment


class TestExperimentTracking(TestEnvVarsCase):
    def setUp(self):
        super(TestExperimentTracking, self).setUp()
        settings.IN_CLUSTER = True

    def test_cluster_def_checks_in_cluster(self):
        settings.IN_CLUSTER = False
        with self.assertRaises(PolyaxonClientException):
            Experiment.get_cluster_def()

    def test_empty_cluster_def(self):
        self.check_empty_value('POLYAXON_CLUSTER', Experiment.get_cluster_def)

    def test_non_dict_cluster_def(self):
        self.check_non_dict_value('POLYAXON_CLUSTER', Experiment.get_cluster_def)

    def test_dict_cluster_def(self):
        cluster_def = {
            "master": ["plxjob-master0-8eefb7a1146f476ca66e3bee9b88c1de:2000"],
            "worker": ["plxjob-worker1-8eefb7a1146f476ca66e3bee9b88c1de:2000",
                       "plxjob-worker2-8eefb7a1146f476ca66e3bee9b88c1de:2000"],
            "ps": ["plxjob-ps3-8eefb7a1146f476ca66e3bee9b88c1de:2000"],
        }
        self.check_valid_dict_value('POLYAXON_CLUSTER', Experiment.get_cluster_def, cluster_def)

    def test_declarations_checks_in_cluster(self):
        settings.IN_CLUSTER = False
        with self.assertRaises(PolyaxonClientException):
            Experiment.get_declarations()

    def test_empty_declarations(self):
        self.check_empty_value('POLYAXON_DECLARATIONS', Experiment.get_declarations)

    def test_non_dict_declarations(self):
        self.check_non_dict_value('POLYAXON_DECLARATIONS', Experiment.get_declarations)

    def test_dict_declarations(self):
        declarations = {
            "foo": "bar"
        }
        self.check_valid_dict_value('POLYAXON_DECLARATIONS',
                                    Experiment.get_declarations,
                                    declarations)

    def test_experiment_info_checks_in_cluster(self):
        settings.IN_CLUSTER = False
        with self.assertRaises(PolyaxonClientException):
            Experiment.get_experiment_info()

    def test_empty_experiment_info(self):
        self.check_empty_value('POLYAXON_EXPERIMENT_INFO', Experiment.get_experiment_info)

    def test_non_dict_experiment_info(self):
        self.check_non_dict_value('POLYAXON_EXPERIMENT_INFO', Experiment.get_experiment_info)

    def test_dict_experiment_info(self):
        experiment_info = {
            "project_name": "project_bar",
            "experiment_group_name": None,
            "experiment_name": "project_bar.1",
            "project_uuid": uuid.uuid4().hex,
            "experiment_group_uuid": None,
            "experiment_uuid": uuid.uuid4().hex,
        }
        self.check_valid_dict_value('POLYAXON_EXPERIMENT_INFO',
                                    Experiment.get_experiment_info,
                                    experiment_info)

    def test_task_info_checks_in_cluster(self):
        settings.IN_CLUSTER = False
        with self.assertRaises(PolyaxonClientException):
            Experiment.get_task_info()

    def test_empty_task_info(self):
        self.check_empty_value('POLYAXON_TASK_INFO', Experiment.get_task_info)

    def test_non_dict_task_info(self):
        self.check_non_dict_value('POLYAXON_TASK_INFO', Experiment.get_task_info)

    def test_dict_task_info(self):
        task_info = {"type": 'master', "index": 0}
        self.check_valid_dict_value('POLYAXON_TASK_INFO',
                                    Experiment.get_task_info,
                                    task_info)

    def test_tf_config_checks_in_cluster(self):
        settings.IN_CLUSTER = False
        with self.assertRaises(PolyaxonClientException):
            Experiment.get_tf_config()

    def test_empty_tf_config(self):
        os.environ['POLYAXON_RUN_OUTPUTS_PATH'] = 'path'
        assert Experiment.get_tf_config() == {
            'cluster': None,
            'task': None,
            'model_dir': 'path',
            'environment': 'cloud'
        }

    def test_non_dict_tf_config(self):
        os.environ['POLYAXON_RUN_OUTPUTS_PATH'] = 'path'
        os.environ['POLYAXON_CLUSTER'] = 'value'
        os.environ['POLYAXON_TASK_INFO'] = 'value'
        assert Experiment.get_tf_config() == {
            'cluster': None,
            'task': None,
            'model_dir': 'path',
            'environment': 'cloud'
        }

    def test_dict_tf_config(self):
        cluster_def = {
            "master": ["plxjob-master0-8eefb7a1146f476ca66e3bee9b88c1de:2000"],
            "worker": ["plxjob-worker1-8eefb7a1146f476ca66e3bee9b88c1de:2000",
                       "plxjob-worker2-8eefb7a1146f476ca66e3bee9b88c1de:2000"],
            "ps": ["plxjob-ps3-8eefb7a1146f476ca66e3bee9b88c1de:2000"],
        }
        task_info = {"type": 'master', "index": 0}
        os.environ['POLYAXON_CLUSTER'] = json.dumps(cluster_def)
        os.environ['POLYAXON_TASK_INFO'] = json.dumps(task_info)
        os.environ['POLYAXON_RUN_OUTPUTS_PATH'] = 'path'
        assert Experiment.get_tf_config() == {
            'cluster': cluster_def,
            'task': {"type": 'master', "index": 0},
            'model_dir': 'path',
            'environment': 'cloud'
        }



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from polyaxon_client.logger import logger
from polyaxon_client.transport.retry_transport import RetryTransportMixin
from polyaxon_client.workers.queue_worker import QueueWorker


class ThreadedTransportMixin(RetryTransportMixin):
    """Threads operations transport."""

    @property
    def threaded_done(self):
        if hasattr(self, '_threaded_done'):
            return self._threaded_done
        return None

    @property
    def threaded_exceptions(self):
        if hasattr(self, '_threaded_exceptions'):
            return self._threaded_exceptions
        return None

    def queue_request(self, request, url, **kwargs):
        try:
            request(url=url, session=self.retry_session, **kwargs)
        except Exception as e:
            self._threaded_exceptions += 1
            logger.debug('Error making request url: %s, params: params %s, exp: %s', url, kwargs, e)
        finally:
            self._threaded_done += 1

    @property
    def worker(self):
        if not hasattr(self, '_worker') or not self._worker.is_alive():
            self._worker = QueueWorker(timeout=self.config.timeout)
            self._worker.start()
        return self._worker

    def async_post(self,
                   url,
                   params=None,
                   data=None,
                   files=None,
                   json_data=None,
                   timeout=None,
                   headers=None):
        """Async Call request with a post."""
        return self.worker.queue(self.queue_request,
                                 request=self.post,
                                 url=url,
                                 params=params,
                                 data=data,
                                 files=files,
                                 json_data=json_data,
                                 timeout=timeout,
                                 headers=headers)

    def async_patch(self,
                    url,
                    params=None,
                    data=None,
                    files=None,
                    json_data=None,
                    timeout=None,
                    headers=None):
        """Async Call request with a patch."""
        return self.worker.queue(self.queue_request,
                                 request=self.patch,
                                 url=url,
                                 params=params,
                                 data=data,
                                 files=files,
                                 json_data=json_data,
                                 timeout=timeout,
                                 headers=headers)

    def async_delete(self,
                     url,
                     params=None,
                     data=None,
                     files=None,
                     json_data=None,
                     timeout=None,
                     headers=None):
        """Async Call request with a delete."""
        return self.worker.queue(self.queue_request,
                                 request=self.delete,
                                 url=url,
                                 params=params,
                                 data=data,
                                 files=files,
                                 json_data=json_data,
                                 timeout=timeout,
                                 headers=headers)

    def async_put(self,
                  url,
                  params=None,
                  data=None,
                  files=None,
                  json_data=None,
                  timeout=None,
                  headers=None):
        """Async Call request with a put."""
        return self.worker.queue(self.queue_request,
                                 request=self.put,
                                 url=url,
                                 params=params,
                                 data=data,
                                 files=files,
                                 json_data=json_data,
                                 timeout=timeout,
                                 headers=headers)

    def async_upload(self,
                     url,
                     files,
                     files_size,
                     params=None,
                     json_data=None,
                     timeout=3600,
                     headers=None):
        return self.worker.queue(self.queue_request,
                                 request=self.upload,
                                 url=url,
                                 files=files,
                                 files_size=files_size,
                                 params=params,
                                 json_data=json_data,
                                 timeout=timeout,
                                 headers=headers)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

import atexit
import json
import os
import sys
import time

from polystores.stores.manager import StoreManager

from polyaxon_client import PolyaxonClient, settings
from polyaxon_client.exceptions import PolyaxonClientException
from polyaxon_client.tracking.in_cluster import ensure_in_custer
from polyaxon_client.tracking.no_op import check_no_op
from polyaxon_client.tracking.paths import get_outputs_path
from polyaxon_client.tracking.utils.project import get_project_info


class BaseTracker(object):
    REQUIRES_OUTPUTS = True

    @check_no_op
    def __init__(self,
                 project=None,
                 client=None,
                 track_logs=True,
                 track_code=True,
                 track_env=True,
                 outputs_store=None):
        if not settings.IN_CLUSTER and project is None:
            raise PolyaxonClientException('Please provide a valid project.')
        elif self.is_notebook_job:
            job_info = self.get_notebook_job_info()
            project = job_info['project_name']

        self.last_status = None
        self.client = client or PolyaxonClient()
        if settings.IN_CLUSTER:
            self.user = None
        else:
            self.user = (self.client.auth.get_user().username
                         if self.client.api_config.schema_response
                         else self.client.auth.get_user().get('username'))

        username, project_name = get_project_info(current_user=self.user, project=project)
        self.track_logs = track_logs
        self.track_code = track_code
        self.track_env = track_env
        self.project = project
        self.username = username
        self.project_name = project_name
        self.outputs_store = outputs_store
        self._data = None

        # Setup the outputs store
        if outputs_store is None and settings.IN_CLUSTER and self.REQUIRES_OUTPUTS:
            self.set_outputs_store(outputs_path=get_outputs_path(), set_env_vars=True)

    @check_no_op
    def get_notebook_job_info(self):
        ensure_in_custer()

        info = os.getenv('POLYAXON_NOTEBOOK_INFO', None)
        try:
            return json.loads(info) if info else None
        except (ValueError, TypeError):
            print('Could get experiment info, '
                  'please make sure this is running inside a polyaxon job.')
            return None

    @property
    def is_notebook_job(self):
        if settings.NO_OP:
            return None

        return settings.IN_CLUSTER and 'POLYAXON_NOTEBOOK_INFO' in os.environ

    def get_data(self):
        raise NotImplementedError

    def _set_health_url(self):
        raise NotImplementedError

    def log_status(self, status, message=None, traceback=None):
        raise NotImplementedError

    @check_no_op
    def _start(self):
        atexit.register(self._end)
        self.start()

        def excepthook(exception, value, tb):
            self.failed(message='Type: {}, Value: {}'.format(exception, value))
            # Resume normal work
            sys.__excepthook__(exception, value, tb)

        sys.excepthook = excepthook

    @check_no_op
    def _end(self):
        self.succeeded()

    @check_no_op
    def start(self):
        self.log_status('running')
        self.last_status = 'running'

    @check_no_op
    def end(self, status, message=None, traceback=None):
        if self.last_status in ['succeeded', 'failed', 'stopped']:
            return
        self.log_status(status=status, message=message, traceback=traceback)
        self.last_status = status
        time.sleep(0.1)  # Just to give the opportunity to the worker to pick the message

    @check_no_op
    def succeeded(self):
        self.end('succeeded')

    @check_no_op
    def stop(self):
        self.end('stopped')

    @check_no_op
    def failed(self, message=None, traceback=None):
        self.end(status='failed', message=message, traceback=traceback)

    @check_no_op
    def set_outputs_store(self, outputs_store=None, outputs_path=None, set_env_vars=False):
        if not any([outputs_store, outputs_path]):
            raise PolyaxonClientException(
                'An Store instance or and outputs path is required.')
        self.outputs_store = outputs_store or StoreManager(path=outputs_path)
        if self.outputs_store and set_env_vars:
            self.outputs_store.set_env_vars()

    @check_no_op
    def log_output(self, filename, **kwargs):
        self.outputs_store.upload_file(filename=filename)

    @check_no_op
    def log_outputs(self, dirname, **kwargs):
        self.outputs_store.upload_dir(dirname=dirname)



from polyaxon_client import settings
from polyaxon_client.exceptions import PolyaxonClientException


def ensure_in_custer():
    if not settings.IN_CLUSTER:
        raise PolyaxonClientException('This experiment/job is not running inside a Polyaxon job.')



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from polyaxon_client.api.base import BaseApiHandler
from polyaxon_client.exceptions import (
    AuthenticationError,
    BadRequestError,
    NotFoundError,
    PolyaxonClientException
)
from polyaxon_client.logger import logger
from polyaxon_client.schemas import DatasetConfig


class DatasetApi(BaseApiHandler):
    """
    Api handler to get datasets from the server.
    """
    ENDPOINT = "/datasets"

    def get_datasets(self, page=1):
        try:
            response = self.transport.get(self._get_http_url())
            return self.prepare_list_results(response.json(), page, DatasetConfig)
        except PolyaxonClientException as e:
            if isinstance(e, AuthenticationError):
                # exit now since there is nothing we can do without login
                raise e
            logger.info("Error while retrieving datasets: %s", e.message)
            return []

    def get_by_name(self, username, datasetname):
        request_url = self.build_url(username, datasetname)
        request_url = self._get_http_url(request_url)
        try:
            response = self.transport.get(request_url)
            return self.prepare_results(response_json=response.json(), config=DatasetConfig)
        except NotFoundError:
            return None

    def create_dataset(self, data):
        """
        Create a temporary directory for the tar file that will be removed at
        the end of the operation.
        """
        try:
            post_body = data.to_dict()
            post_body["resumable"] = True
            response = self.transport.post(self._get_http_url(), json_data=post_body)
            return response.json()
        except BadRequestError as e:
            logger.error('Could not create data %s', e.message)
            return None
        except PolyaxonClientException as e:
            logger.error('Could not create data %s', e.message)
            return None

    def delete_dataset(self, data_uuid, background=False):
        request_url = self._get_http_url(data_uuid)

        if background:
            self.transport.async_delete(request_url, timeout=60)
            return None

        try:
            return self.transport.delete(request_url, timeout=60)
        except PolyaxonClientException as e:
            logger.error('Could not create data %s', e.message)
            return False



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

import getpass
import os
import platform
import socket
import sys

from polyaxon_client.logger import logger

PROJECT_CLIENT_NAME = 'polyaxon-client'


def is_notebook():
    return 'ipykernel' in sys.modules


def get_filename():
    if is_notebook():
        return 'notebook'
    try:
        return os.path.basename(__file__)
    except Exception as e:
        logger.debug('Could not detect filename, %s', e)
        return 'not found'


def get_module_path():
    try:
        return os.path.dirname(os.path.realpath('__file__'))
    except Exception as e:
        logger.debug('Could not detect module path, %s', e)
        return 'not found'


def get_run_env():
    import pkg_resources

    def get_packages():
        try:
            installed_packages = [d for d in pkg_resources.working_set]  # noqa
            return sorted(["{}=={}".format(pkg.key, pkg.version) for pkg in installed_packages])
        except Exception as e:
            logger.debug('Could not detect installed packages, %s', e)
            return []

    try:
        version = pkg_resources.get_distribution(PROJECT_CLIENT_NAME).version
    except pkg_resources.DistributionNotFound:
        version = ''
    return {
        'pid': os.getpid(),
        'hostname': socket.gethostname(),
        'os': platform.platform(aliased=True),
        'system': platform.system(),
        'python_version_verbose': sys.version,
        'python_version': platform.python_version(),
        'user': getpass.getuser(),
        'client_version': version,
        'sys.argv': sys.argv,
        'is_notebook': is_notebook(),
        'filename': get_filename(),
        'module_path': get_module_path(),
        'packages': get_packages(),
    }



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from polyaxon_client.exceptions import PolyaxonClientException

try:
    from keras.callbacks import Callback
except ImportError:
    try:
        from tensorflow.python.keras.callbacks import Callback
    except ImportError:
        raise PolyaxonClientException('Keras is required to use PolyaxonKeras')


class PolyaxonKeras(Callback):

    def __init__(self, experiment, metrics=None):
        self.experiment = experiment
        self.metrics = metrics

    def on_epoch_end(self, epoch, logs=None):
        if not logs:
            return
        if self.metrics:
            metrics = {metric: logs[metric] for metric in self.metrics if metric in logs}
        else:
            metrics = logs  # Log all metrics

        self.experiment.log_metrics(**metrics)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from hestia.auth import AuthenticationTypes
from tests.test_transports.utils import BaseTestCaseTransport

from polyaxon_client.api_config import ApiConfig
from polyaxon_client.transport import Transport


class TestTransport(BaseTestCaseTransport):
    # pylint:disable=protected-access
    def setUp(self):
        super(TestTransport, self).setUp()
        self.transport = Transport()

    def test_get_headers(self):
        assert self.transport._get_headers() == {}
        assert self.transport._get_headers({'foo': 'bar'}) == {'foo': 'bar'}

        self.transport.config = ApiConfig(token='token', host='host')

        assert self.transport._get_headers() == {
            'Authorization': "{} {}".format(AuthenticationTypes.TOKEN, 'token')
        }
        assert self.transport._get_headers({'foo': 'bar'}) == {
            'foo': 'bar',
            'Authorization': "{} {}".format(AuthenticationTypes.TOKEN, 'token')
        }

        self.transport.config.authentication_type = AuthenticationTypes.INTERNAL_TOKEN
        assert self.transport._get_headers() == {
            'Authorization': "{} {}".format(AuthenticationTypes.INTERNAL_TOKEN, 'token')
        }
        assert self.transport._get_headers({'foo': 'bar'}) == {
            'foo': 'bar',
            'Authorization': "{} {}".format(AuthenticationTypes.INTERNAL_TOKEN, 'token')
        }



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from polyaxon_client import settings
from polyaxon_client.api.auth import AuthApi
from polyaxon_client.api.bookmark import BookmarkApi
from polyaxon_client.api.build_job import BuildJobApi
from polyaxon_client.api.cluster import ClusterApi
from polyaxon_client.api.experiment import ExperimentApi
from polyaxon_client.api.experiment_group import ExperimentGroupApi
from polyaxon_client.api.experiment_job import ExperimentJobApi
from polyaxon_client.api.job import JobApi
from polyaxon_client.api.project import ProjectApi
from polyaxon_client.api.user import UserApi
from polyaxon_client.api.version import VersionApi
from polyaxon_client.api_config import ApiConfig
from polyaxon_client.transport import Transport


class PolyaxonClient(object):
    def __init__(self,
                 api_config=None,
                 host=None,
                 token=None,
                 http_port=None,
                 ws_port=None,
                 use_https=False,
                 verify_ssl=None,
                 in_cluster=None,
                 authentication_type=None,
                 api_version=None,
                 reraise=False,
                 schema_response=None,
                 timeout=None):

        self._api_config = api_config or ApiConfig(host=host,
                                                   http_port=http_port,
                                                   ws_port=ws_port,
                                                   token=token,
                                                   authentication_type=authentication_type,
                                                   version=api_version,
                                                   use_https=use_https,
                                                   verify_ssl=verify_ssl,
                                                   in_cluster=in_cluster,
                                                   schema_response=schema_response,
                                                   reraise=reraise,
                                                   timeout=timeout)

        self._transport = None
        self._auth_api = None
        self._cluster_api = None
        self._version_api = None
        self._project_api = None
        self._experiment_group_api = None
        self._experiment_api = None
        self._experiment_job_api = None
        self._job_api = None
        self._build_job_api = None
        self._user_api = None
        self._bookmark_api = None

    def reset(self):
        self._transport = None
        self._auth_api = None
        self._cluster_api = None
        self._version_api = None
        self._project_api = None
        self._experiment_group_api = None
        self._experiment_api = None
        self._experiment_job_api = None
        self._job_api = None
        self._build_job_api = None
        self._user_api = None
        self._bookmark_api = None

    @property
    def host(self):
        return self.api_config.host

    @property
    def http_port(self):
        return self.api_config.http_port

    @property
    def ws_port(self):
        return self.api_config.ws_port

    @property
    def use_https(self):
        return self.api_config.use_https

    @property
    def verify_ssl(self):
        return self.api_config.verify_ssl

    @property
    def token(self):
        return self.api_config.token

    @property
    def authentication_type(self):
        return self.api_config.authentication_type

    @property
    def in_cluster(self):
        return self.api_config.in_cluster

    @property
    def api_version(self):
        return self.api_config.version

    @property
    def reraise(self):
        return self.api_config.reraise

    @property
    def timeout(self):
        return self.api_config.timeout

    def set_host(self, host):
        self.api_config.host = host
        self.reset()

    def set_http_port(self, http_port):
        self.api_config.http_port = http_port
        self.reset()

    def set_ws_port(self, ws_port):
        self.api_config.ws_port = ws_port
        self.reset()

    def set_use_https(self, use_https):
        self.api_config.use_https = use_https
        self.reset()

    def set_verify_ssl(self, verify_ssl):
        self.api_config.verify_ssl = verify_ssl
        self.reset()

    def set_token(self, token):
        self.api_config.token = token
        self.reset()

    def set_in_cluster(self, in_cluster):
        self.api_config.in_cluster = in_cluster
        self.reset()

    def set_authentication_type(self, authentication_type):
        self.api_config.authentication_type = authentication_type
        self.reset()

    def set_version_api(self, version_api):
        self.api_config.version = version_api
        self.reset()

    def set_reraise(self, reraise):
        self.api_config.reraise = reraise
        self.reset()

    def set_health_check(self, url):
        self.transport.set_health_check(url)

    def set_internal_health_check(self):
        if settings.INTERNAL_HEALTH_CHECK_URL:
            self.set_health_check(
                self.auth.build_url(self.api_config.base_url, settings.INTERNAL_HEALTH_CHECK_URL))

    @property
    def transport(self):
        if not self._transport:
            self._transport = Transport(config=self.api_config)
        return self._transport

    @property
    def api_config(self):
        return self._api_config

    @property
    def auth(self):
        if not self._auth_api:
            self._auth_api = AuthApi(transport=self.transport,
                                     config=self.api_config)
        return self._auth_api

    @property
    def cluster(self):
        if not self._cluster_api:
            self._cluster_api = ClusterApi(transport=self.transport,
                                           config=self.api_config)
        return self._cluster_api

    @property
    def version(self):
        if not self._version_api:
            self._version_api = VersionApi(transport=self.transport,
                                           config=self.api_config)
        return self._version_api

    @property
    def project(self):
        if not self._project_api:
            self._project_api = ProjectApi(transport=self.transport,
                                           config=self.api_config)
        return self._project_api

    @property
    def experiment_group(self):
        if not self._experiment_group_api:
            self._experiment_group_api = ExperimentGroupApi(transport=self.transport,
                                                            config=self.api_config)
        return self._experiment_group_api

    @property
    def experiment(self):
        if not self._experiment_api:
            self._experiment_api = ExperimentApi(transport=self.transport,
                                                 config=self.api_config)
        return self._experiment_api

    @property
    def experiment_job(self):
        if not self._experiment_job_api:
            self._experiment_job_api = ExperimentJobApi(transport=self.transport,
                                                        config=self.api_config)
        return self._experiment_job_api

    @property
    def job(self):
        if not self._job_api:
            self._job_api = JobApi(transport=self.transport,
                                   config=self.api_config)
        return self._job_api

    @property
    def build_job(self):
        if not self._build_job_api:
            self._build_job_api = BuildJobApi(transport=self.transport,
                                              config=self.api_config)
        return self._build_job_api

    @property
    def user(self):
        if not self._user_api:
            self._user_api = UserApi(transport=self.transport,
                                     config=self.api_config)
        return self._user_api

    @property
    def bookmark(self):
        if not self._bookmark_api:
            self._bookmark_api = BookmarkApi(transport=self.transport,
                                             config=self.api_config)
        return self._bookmark_api



import logging

from polyaxon_client import settings
from polyaxon_client.handlers.handler import PolyaxonHandler

EXCLUDE_DEFAULT_LOGGERS = (
    'polyaxon.client',
    'polyaxon.cli',
)


def setup_logging(send_logs, exclude=EXCLUDE_DEFAULT_LOGGERS):
    if settings.IN_CLUSTER:
        return

    logger = logging.getLogger()
    logger.setLevel(logging.INFO)
    logger.propagate = False
    if PolyaxonHandler in map(type, logger.handlers):
        for handler in logger.handlers:
            if isinstance(handler, PolyaxonHandler):
                handler.set_send_logs(send_logs=send_logs)
    else:
        handler = PolyaxonHandler(send_logs=send_logs)

    logger.addHandler(handler)

    for logger_name in exclude:
        logger = logging.getLogger(logger_name)
        logger.propagate = False
        logger.addHandler(logging.StreamHandler())



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

import json
import os

from polyaxon_client.tracking.in_cluster import ensure_in_custer
from polyaxon_client.tracking.no_op import check_no_op


@check_no_op
def get_data_paths():
    """
    The data paths dictionary that you mounted for the job/experiment.
    {'data1': '/data/1/', 'data-foo': '/data/foo'}
    """
    ensure_in_custer()
    data_path = os.getenv('POLYAXON_RUN_DATA_PATHS', None)
    try:
        return json.loads(data_path) if data_path else None
    except (ValueError, TypeError):
        print('Could get data paths, '
              'please make sure this is running inside a polyaxon job.')
        return None


@check_no_op
def get_base_outputs_path():
    """
    The bae outputs path generated
    """
    return os.getenv('POLYAXON_BASE_OUTPUTS_PATH', None)


@check_no_op
def get_outputs_path():
    """
    The outputs path generated by polyaxon based on the hierarchy of the experiment:
    Experiment:
        `user/project/group/experiment/files`
    Jobs:
        `user/project/jobs/files`
    """
    ensure_in_custer()
    return os.getenv('POLYAXON_RUN_OUTPUTS_PATH', None)


@check_no_op
def get_outputs_refs_paths():
    """
    The references outputs paths requested by the user,
    the order follows the order specified by the user:
    {
        'jobs': [
            `user/project/job12/files`,
        ],
        'experiments': [
            `user/project/group/experiment1/files`,
            `user/project/experiment100/files`
        ]
    }
    """
    ensure_in_custer()
    outputs_refs = os.getenv('POLYAXON_REFS_OUTPUTS_PATHS', None)
    try:
        return json.loads(outputs_refs) if outputs_refs else None
    except (ValueError, TypeError):
        print('Could get outputs refs paths, '
              'please make sure this is running inside a polyaxon job.')
        return None


@check_no_op
def get_log_level():
    """If set on the polyaxonfile it will return the log level."""
    ensure_in_custer()
    return os.getenv('POLYAXON_LOG_LEVEL', None)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

import requests

from requests.adapters import HTTPAdapter

from urllib3 import Retry


class RetryTransportMixin(object):
    """Threads operations transport."""

    @property
    def retry_session(self):
        if not hasattr(self, '_retry_session'):
            self._retry_session = requests.Session()
            retry = Retry(
                total=3,
                read=3,
                connect=3,
                backoff_factor=2,
                status_forcelist=[429, 500, 502, 503, 504],
            )
            adapter = HTTPAdapter(max_retries=retry)
            self._retry_session.mount('http://', adapter)
            self._retry_session.mount('https://', adapter)
            self._threaded_done = 0
            self._threaded_exceptions = 0
            self._periodic_http_done = 0
            self._periodic_http_exceptions = 0
            self._periodic_ws_done = 0
            self._periodic_ws_exceptions = 0
        return self._retry_session
